import hashlib
import itertools
import os
from datetime import datetime, timezone
from io import BytesIO
from typing import Optional

import pytest
import tzlocal
import yaml
from asn1crypto import algos, cms, core, keys
from asn1crypto.algos import (
    DigestAlgorithm,
    MaskGenAlgorithm,
    RSASSAPSSParams,
    SignedDigestAlgorithm,
)
from certomancer import PKIArchitecture
from certomancer.registry import ArchLabel, CertLabel, KeyLabel
from freezegun import freeze_time
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign import attributes, fields, signers, timestamps
from pyhanko.sign.ades.api import CAdESSignedAttrSpec, SignerAttrSpec
from pyhanko.sign.ades.report import AdESIndeterminate, AdESStatus
from pyhanko.sign.attributes import (
    CMSAttributeProvider,
    SignedAttributeProviderSpec,
    TSTProvider,
    UnsignedAttributeProviderSpec,
)
from pyhanko.sign.general import (
    CMSExtractionError,
    CMSStructuralError,
    SigningError,
    as_signing_certificate,
    as_signing_certificate_v2,
    find_cms_attribute,
    simple_cms_attribute,
)
from pyhanko.sign.signers import cms_embedder
from pyhanko.sign.signers.pdf_cms import (
    PdfCMSSignedAttributes,
    select_suitable_signing_md,
)
from pyhanko.sign.validation import (
    DocumentSecurityStore,
    StandardCMSSignatureStatus,
    async_validate_cms_signature,
    async_validate_detached_cms,
    async_validate_pdf_ltv_signature,
    async_validate_pdf_signature,
    collect_validation_info,
    validate_cms_signature,
)
from pyhanko.sign.validation.errors import (
    DisallowedAlgorithmError,
    SignatureValidationError,
)
from pyhanko.sign.validation.generic_cms import validate_sig_integrity
from pyhanko.sign.validation.status import ClaimedAttributes
from pyhanko.sign.validation.utils import CMSAlgorithmUsagePolicy

from pyhanko_certvalidator import ValidationContext
from pyhanko_certvalidator.policy_decl import (
    AlgorithmUsageConstraint,
    DisallowWeakAlgorithmsPolicy,
)
from pyhanko_certvalidator.registry import SimpleCertificateStore

from .samples import (
    CERTOMANCER,
    CRYPTO_DATA_DIR,
    MINIMAL,
    PDF_DATA_DIR,
    SAMPLE_GROUP_ATTR,
    TESTING_CA,
    TESTING_CA_DSA,
    TESTING_CA_ECDSA,
    TESTING_CA_ED448,
    TESTING_CA_ED25519,
    TESTING_CA_ERRORS,
)
from .signing_commons import (
    DSA_INTERM_CERT,
    DSA_ROOT_CERT,
    DUMMY_TS,
    ECC_INTERM_CERT,
    ECC_ROOT_CERT,
    FIXED_OCSP,
    FROM_CA,
    FROM_DSA_CA,
    FROM_ECC_CA,
    INTERM_CERT,
    ROOT_CERT,
    SIMPLE_DSA_V_CONTEXT,
    SIMPLE_ECC_V_CONTEXT,
    SIMPLE_V_CONTEXT,
    async_val_trusted,
    live_ac_vcs,
    live_testing_vc,
    val_trusted,
    val_untrusted,
)


def test_generic_data_sign_legacy():
    input_data = b'Hello world!'
    with pytest.deprecated_call():
        # noinspection PyDeprecation
        signature = FROM_CA.sign_general_data(
            input_data, 'sha256', detached=True
        )

    # reset the stream
    if isinstance(input_data, BytesIO):
        input_data.seek(0)

    # re-parse just to make sure we're starting fresh
    signature = cms.ContentInfo.load(signature.dump())

    raw_digest = hashlib.sha256(b'Hello world!').digest()
    content = signature['content']
    assert content['version'].native == 'v1'
    assert isinstance(content, cms.SignedData)

    with pytest.deprecated_call():
        # noinspection PyDeprecation

        # noinspection PyDeprecation
        status = validate_cms_signature(content, raw_digest=raw_digest)
    assert status.valid
    assert status.intact

    eci = content['encap_content_info']
    assert eci['content_type'].native == 'data'
    assert eci['content'].native is None

    assert status.valid
    assert status.intact


@pytest.mark.parametrize(
    'input_data, detached',
    list(
        itertools.product(
            [
                b'Hello world!',
                BytesIO(b'Hello world!'),
                # v1 CMS -> PKCS#7 compatible -> use cms.ContentInfo
                cms.ContentInfo(
                    {'content_type': 'data', 'content': b'Hello world!'}
                ),
            ],
            [True, False],
        )
    ),
)
@pytest.mark.asyncio
async def test_generic_data_sign(input_data, detached):
    signature = await FROM_CA.async_sign_general_data(
        input_data, 'sha256', detached=detached
    )

    # reset the stream
    if isinstance(input_data, BytesIO):
        input_data.seek(0)

    # re-parse just to make sure we're starting fresh
    signature = cms.ContentInfo.load(signature.dump())

    raw_digest = hashlib.sha256(b'Hello world!').digest() if detached else None
    content = signature['content']
    assert content['version'].native == 'v1'
    assert isinstance(content, cms.SignedData)
    status = await async_validate_cms_signature(content, raw_digest=raw_digest)
    assert status.valid
    assert status.intact

    eci = content['encap_content_info']
    if detached:
        assert eci['content_type'].native == 'data'
        assert eci['content'].native is None

        status = await async_validate_detached_cms(input_data, content)
        assert status.valid
        assert status.intact
        assert (
            'No available information about the signing time.'
            in status.pretty_print_details()
        )
        if isinstance(input_data, BytesIO):
            input_data.seek(0)
    else:
        assert eci['content_type'].native == 'data'
        assert eci['content'].native == b'Hello world!'

    assert status.valid
    assert status.intact

    assert content['signer_infos'][0]['unsigned_attrs'].native is None


@pytest.mark.parametrize('detached', [True, False])
@pytest.mark.asyncio
async def test_cms_v3_sign(detached):
    inner_obj = await FROM_CA.async_sign_general_data(
        b'Hello world!', 'sha256', detached=False
    )

    signature = await FROM_CA.async_sign_general_data(
        cms.EncapsulatedContentInfo(
            {
                'content_type': 'signed_data',
                'content': inner_obj['content'].untag(),
            }
        ),
        'sha256',
        detached=detached,
    )

    # re-parse just to make sure we're starting fresh
    signature = cms.ContentInfo.load(signature.dump())

    content = signature['content']
    assert content['version'].native == 'v3'
    assert isinstance(content, cms.SignedData)
    eci = content['encap_content_info']
    assert eci['content_type'].native == 'signed_data'
    if detached:
        raw_digest = hashlib.sha256(
            inner_obj['content'].untag().dump()
        ).digest()
    else:
        raw_digest = None
        inner_eci = eci['content'].parsed['encap_content_info']
        assert inner_eci['content'].native == b'Hello world!'
    status = await async_validate_cms_signature(content, raw_digest=raw_digest)
    assert status.valid
    assert status.intact


@pytest.mark.asyncio
async def test_detached_cms_with_self_reported_timestamp():
    dt = datetime.fromisoformat('2020-11-01T05:00:00+00:00')
    signature = await FROM_CA.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
        signed_attr_settings=PdfCMSSignedAttributes(signing_time=dt),
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!', signature['content']
    )
    assert status.signer_reported_dt == dt
    assert status.timestamp_validity is None
    assert 'reported by signer' in status.pretty_print_details()
    assert status.valid
    assert status.intact


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_tst():
    signature = await FROM_CA.async_sign_general_data(
        b'Hello world!', 'sha256', detached=True, timestamper=DUMMY_TS
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!', signature['content']
    )
    assert status.signer_reported_dt is None
    assert status.timestamp_validity.intact
    assert status.timestamp_validity.valid
    assert status.timestamp_validity.timestamp == datetime.now(tz=timezone.utc)
    assert 'The TSA certificate is untrusted' in status.pretty_print_details()
    assert status.valid
    assert status.intact


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_content_tst():
    signed_attr_settings = PdfCMSSignedAttributes(
        cades_signed_attrs=CAdESSignedAttrSpec(timestamp_content=True)
    )
    signature = await FROM_CA.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
        timestamper=DUMMY_TS,
        signed_attr_settings=signed_attr_settings,
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!', signature['content']
    )
    assert status.signer_reported_dt is None
    assert status.timestamp_validity.intact
    assert status.timestamp_validity.valid
    assert status.timestamp_validity.timestamp == datetime.now(tz=timezone.utc)
    assert status.content_timestamp_validity
    assert status.content_timestamp_validity.intact
    assert status.content_timestamp_validity.valid
    assert status.content_timestamp_validity.timestamp == datetime.now(
        tz=timezone.utc
    )
    pretty_print = status.pretty_print_details()
    assert 'The TSA certificate is untrusted' in pretty_print
    assert 'Content timestamp' in pretty_print
    assert 'Signature timestamp' in pretty_print
    assert status.valid
    assert status.intact
    assert 'CONTENT_TIMESTAMP_TOKEN<INTACT:UNTRUSTED>' in status.summary()
    assert ',TIMESTAMP_TOKEN<INTACT:UNTRUSTED>' in status.summary()


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_wrong_tst():
    class Spec(UnsignedAttributeProviderSpec):
        def unsigned_attr_providers(
            self,
            signature: bytes,
            signed_attrs: cms.CMSAttributes,
            digest_algorithm: str,
        ):
            yield TSTProvider(
                digest_algorithm='sha256',
                data_to_ts=b'\xde\xad\xbe\xef',
                timestamper=DUMMY_TS,
            )

    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    signer.unsigned_attr_prov_spec = Spec()
    signature = await signer.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
        timestamper=DUMMY_TS,
        signed_attr_settings=PdfCMSSignedAttributes(
            cades_signed_attrs=CAdESSignedAttrSpec(timestamp_content=True)
        ),
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!', signature['content']
    )
    assert status.signer_reported_dt is None
    # signature in TS is valid, but digest is wrong
    assert not status.timestamp_validity.intact
    assert status.timestamp_validity.valid

    assert status.timestamp_validity.timestamp == datetime.now(tz=timezone.utc)
    assert status.content_timestamp_validity
    assert status.content_timestamp_validity.intact
    assert status.content_timestamp_validity.valid
    assert status.content_timestamp_validity.timestamp == datetime.now(
        tz=timezone.utc
    )
    assert status.valid
    assert status.intact
    assert 'CONTENT_TIMESTAMP_TOKEN<INTACT:UNTRUSTED>' in status.summary()
    assert 'TIMESTAMP_TOKEN<INVALID>' in status.summary()


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_wrong_content_tst():
    class Spec(SignedAttributeProviderSpec):
        def signed_attr_providers(
            self, data_digest: bytes, digest_algorithm: str
        ):
            yield TSTProvider(
                digest_algorithm='sha256',
                data_to_ts=b'\xde\xad\xbe\xef',
                timestamper=DUMMY_TS,
                attr_type='content_time_stamp',
            )

    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    signer.signed_attr_prov_spec = Spec()
    signature = await signer.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
        timestamper=DUMMY_TS,
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!', signature['content']
    )
    assert status.signer_reported_dt is None
    assert status.timestamp_validity.intact
    assert status.timestamp_validity.valid
    assert status.timestamp_validity.timestamp == datetime.now(tz=timezone.utc)
    assert status.content_timestamp_validity
    assert not status.content_timestamp_validity.intact
    # internally, the content timestamp is OK
    assert status.content_timestamp_validity.valid
    assert status.content_timestamp_validity.timestamp == datetime.now(
        tz=timezone.utc
    )
    pretty_print = status.pretty_print_details()
    assert 'The TSA certificate is untrusted' in pretty_print
    assert 'Content timestamp' in pretty_print
    assert 'Signature timestamp' in pretty_print
    assert status.valid
    assert status.intact
    assert 'CONTENT_TIMESTAMP_TOKEN<INVALID>' in status.summary()
    assert 'TIMESTAMP_TOKEN<INTACT:UNTRUSTED>' in status.summary()


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_duplicated_attr():
    class CustomProvider(CMSAttributeProvider):
        attribute_type = 'message_digest'

        async def build_attr_value(self, dry_run=False):
            return b'\xde\xad\xbe\xef'

    class Spec(SignedAttributeProviderSpec):
        def signed_attr_providers(
            self, data_digest: bytes, digest_algorithm: str
        ):
            yield CustomProvider()

    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    signer.signed_attr_prov_spec = Spec()
    signature = await signer.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=False,
        timestamper=DUMMY_TS,
    )
    signature = cms.ContentInfo.load(signature.dump())
    with pytest.raises(SignatureValidationError, match='structural.*duplicate'):
        await async_validate_cms_signature(
            signature['content'],
        )


@freeze_time('2020-11-01')
@pytest.mark.parametrize(
    'content,detach',
    [
        (b'This is not a TST!', True),
        (b'This is not a TST!', False),
        (
            cms.ContentInfo(
                {'content_type': 'data', 'content': b'This is not a TST!'}
            ),
            False,
        ),
        (
            cms.EncapsulatedContentInfo(
                {
                    'content_type': '2.999',
                    'content': core.ParsableOctetString(
                        core.OctetString(b'This is not a TST!').dump()
                    ),
                }
            ),
            False,
        ),
        (
            cms.ContentInfo(
                {
                    'content_type': '2.999',
                }
            ),
            True,
        ),
    ],
)
@pytest.mark.asyncio
async def test_detached_with_malformed_content_tst(content, detach):
    class CustomProvider(CMSAttributeProvider):
        attribute_type = 'content_time_stamp'

        async def build_attr_value(self, dry_run=False):
            attr_value = await FROM_CA.async_sign_general_data(
                content,
                'sha256',
                detached=detach,
            )
            return attr_value

    class CustomSigner(signers.SimpleSigner):
        def _signed_attr_providers(self, *args, **kwargs):
            yield from super()._signed_attr_providers(*args, **kwargs)
            yield CustomProvider()

    signer = CustomSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    signature = await signer.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=False,
        timestamper=DUMMY_TS,
    )
    signature = cms.ContentInfo.load(signature.dump())
    with pytest.raises(
        SignatureValidationError, match="does not encapsulate TSTInfo"
    ):
        await async_validate_detached_cms(b'Hello world!', signature['content'])


@freeze_time('2020-11-01')
def test_overspecify_cms_digest_algo():
    # TODO this behaviour is not ideal, but at least this test documents it

    signer = signers.SimpleSigner.load(
        CRYPTO_DATA_DIR + '/selfsigned.key.pem',
        CRYPTO_DATA_DIR + '/selfsigned.cert.pem',
        ca_chain_files=(CRYPTO_DATA_DIR + '/selfsigned.cert.pem',),
        key_passphrase=b'secret',
        # specify an algorithm object that also mandates a specific
        # message digest
        signature_mechanism=SignedDigestAlgorithm({'algorithm': 'sha256_rsa'}),
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    # digest methods agree, so that should be OK
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1', md_algorithm='sha256'),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_untrusted(s)

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with pytest.raises(SigningError):
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1', md_algorithm='sha512'
            ),
            signer=signer,
        )


def _tamper_with_signed_attrs(
    attr_name,
    *,
    duplicate=False,
    oid_duplicate=False,
    delete=False,
    replace_with=None,
    resign=False,
):
    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)
    md_algorithm = 'sha256'

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)
    sig_obj = signers.SignatureObject(bytes_reserved=8192)

    cms_writer.send(cms_embedder.SigObjSetup(sig_placeholder=sig_obj))

    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm, in_place=True)
    )

    signer: signers.SimpleSigner = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    with pytest.deprecated_call():
        # noinspection PyDeprecation
        cms_obj = signer.sign(
            data_digest=prep_digest.document_digest,
            digest_algorithm=md_algorithm,
        )
    sd = cms_obj['content']
    (si,) = sd['signer_infos']
    signed_attrs = si['signed_attrs']
    ix = next(
        ix
        for ix, attr in enumerate(signed_attrs)
        if attr['type'].native == attr_name
    )

    # mess with the attribute in the requested way
    if delete:
        del signed_attrs[ix]
    elif duplicate:
        vals = signed_attrs[ix]['values']
        vals.append(vals[0])
    elif oid_duplicate:
        signed_attrs.append(signed_attrs[ix])
    else:
        vals = signed_attrs[ix]['values']
        vals[0] = replace_with

    # ... and replace the signature if requested
    if resign:
        si['signature'] = signer.sign_raw(
            si['signed_attrs'].untag().dump(), md_algorithm
        )
    cms_writer.send(cms_obj)
    return output


@pytest.mark.parametrize(
    'replacement_value',
    [
        cms.CMSAlgorithmProtection(
            {
                'digest_algorithm': DigestAlgorithm({'algorithm': 'sha1'}),
                'signature_algorithm': SignedDigestAlgorithm(
                    {'algorithm': 'rsassa_pkcs1v15'}
                ),
            }
        ),
        cms.CMSAlgorithmProtection(
            {
                'digest_algorithm': DigestAlgorithm({'algorithm': 'sha256'}),
                'signature_algorithm': SignedDigestAlgorithm(
                    {'algorithm': 'sha512_rsa'}
                ),
            }
        ),
        cms.CMSAlgorithmProtection(
            {
                'digest_algorithm': DigestAlgorithm({'algorithm': 'sha256'}),
            }
        ),
        None,
    ],
)
def test_cms_algorithm_protection(replacement_value):
    output = _tamper_with_signed_attrs(
        'cms_algorithm_protection',
        duplicate=replacement_value is None,
        replace_with=replacement_value,
        resign=True,
    )

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()
    with pytest.raises(SignatureValidationError, match='.*CMS.*'):
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)


def test_signed_attrs_tampering():
    # delete the (signed) CMSAlgorithmProtection attribute
    # this should invalidate the signature

    output = _tamper_with_signed_attrs('cms_algorithm_protection', delete=True)

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()

    intact, valid = validate_sig_integrity(
        emb.signer_info, emb.signer_cert, 'data', digest
    )
    # "intact" refers to the messageDigest attribute, which we didn't touch
    assert intact and not valid


def test_no_message_digest():
    output = _tamper_with_signed_attrs(
        'message_digest', delete=True, resign=True
    )

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()

    with pytest.raises(SignatureValidationError):
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)


def test_duplicate_content_type():
    output = _tamper_with_signed_attrs(
        'content_type', duplicate=True, resign=True
    )

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()

    with pytest.raises(SignatureValidationError) as exc_info:
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)
    assert exc_info.value.ades_status == AdESStatus.FAILED


def test_duplicate_content_type_oid():
    output = _tamper_with_signed_attrs(
        'content_type', oid_duplicate=True, resign=True
    )

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()

    with pytest.raises(CMSStructuralError):
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)


def test_no_content_type():
    output = _tamper_with_signed_attrs('content_type', delete=True, resign=True)

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()

    with pytest.raises(SignatureValidationError):
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)


def test_wrong_content_type():
    # delete the (signed) CMSAlgorithmProtection attribute
    # this should invalidate the signature

    output = _tamper_with_signed_attrs(
        'content_type', replace_with='enveloped_data', resign=True
    )

    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()

    with pytest.raises(SignatureValidationError):
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)


@freeze_time('2020-11-01')
def test_sign_weak_digest():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1', md_algorithm='md5')
    out = signers.sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    with pytest.raises(DisallowedAlgorithmError):
        val_trusted(emb)

    lenient_vc = ValidationContext(
        trust_roots=[ROOT_CERT], weak_hash_algos=set()
    )
    val_trusted(emb, vc=lenient_vc)


@freeze_time('2020-11-01')
def test_sign_weak_digest_prevention():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        md_algorithm='md5',
        validation_context=SIMPLE_V_CONTEXT(),
    )
    with pytest.raises(SigningError, match='md5 is not allowed'):
        signers.sign_pdf(w, meta, signer=FROM_CA)


@freeze_time('2020-11-01')
def test_forbidden_signature_algorithm():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    rsa_banned_vc = ValidationContext(
        trust_roots=[ROOT_CERT],
        algorithm_usage_policy=DisallowWeakAlgorithmsPolicy(
            weak_signature_algos={'rsassa_pkcs1v15'}
        ),
    )
    with pytest.raises(DisallowedAlgorithmError, match="rsa"):
        val_trusted(r.embedded_signatures[0], vc=rsa_banned_vc)


async def _generate_sig_with_mismatching_digest_algos():
    # We have to jump through some hoops to put together a signature
    # where the signing method's digest is not the same as the "external"
    # digest. This is intentional, since it's bad practice.

    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)

    timestamp = datetime.now(tz=tzlocal.get_localzone())
    sig_obj = signers.SignatureObject(timestamp=timestamp, bytes_reserved=8192)

    external_md_algorithm = 'sha256'
    cms_writer.send(cms_embedder.SigObjSetup(sig_placeholder=sig_obj))

    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(
            md_algorithm=external_md_algorithm, in_place=True
        )
    )
    signer = signers.SimpleSigner(
        signing_cert=TESTING_CA.get_cert(CertLabel('signer1')),
        signing_key=TESTING_CA.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [ROOT_CERT, INTERM_CERT]
        ),
    )
    cms_obj = await signer.async_sign(
        data_digest=prep_digest.document_digest,
        digest_algorithm=external_md_algorithm,
        signed_attr_settings=PdfCMSSignedAttributes(signing_time=timestamp),
    )
    si_obj: cms.SignerInfo = cms_obj['content']['signer_infos'][0]
    bad_algo = SignedDigestAlgorithm({'algorithm': 'md5_rsa'})
    si_obj['signature_algorithm'] = signer._signature_mechanism = bad_algo
    attrs = si_obj['signed_attrs']
    cms_prot = find_cms_attribute(attrs, 'cms_algorithm_protection')[0]
    cms_prot['signature_algorithm'] = bad_algo
    # recompute the signature
    si_obj['signature'] = signer.sign_raw(attrs.untag().dump(), 'md5')
    cms_writer.send(cms_obj)
    return output


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_weak_sig_digest_mismatch():
    output = await _generate_sig_with_mismatching_digest_algos()
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]

    lenient_vc = ValidationContext(
        trust_roots=[ROOT_CERT], weak_hash_algos=set()
    )
    with pytest.raises(
        SignatureValidationError, match="sha256 does not match.*md5"
    ):
        await async_val_trusted(emb, vc=lenient_vc)


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_weak_sig_digest_mismatch_allowed():
    output = await _generate_sig_with_mismatching_digest_algos()
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]

    class _AllowEverything(CMSAlgorithmUsagePolicy):
        def digest_combination_allowed(
            self,
            signature_algo: algos.SignedDigestAlgorithm,
            message_digest_algo: algos.DigestAlgorithm,
            moment: Optional[datetime],
        ) -> AlgorithmUsageConstraint:
            return AlgorithmUsageConstraint(allowed=True)

        def digest_algorithm_allowed(
            self, algo: algos.DigestAlgorithm, moment: Optional[datetime]
        ) -> AlgorithmUsageConstraint:
            return AlgorithmUsageConstraint(allowed=True)

        def signature_algorithm_allowed(
            self,
            algo: algos.SignedDigestAlgorithm,
            moment: Optional[datetime],
            public_key: Optional[keys.PublicKeyInfo],
        ) -> AlgorithmUsageConstraint:
            return AlgorithmUsageConstraint(allowed=True)

    val_status = await async_validate_pdf_signature(
        emb,
        SIMPLE_V_CONTEXT(),
        skip_diff=True,
        algorithm_policy=_AllowEverything(),
    )
    assert val_status.valid and val_status.intact


class _AllowMismatches(CMSAlgorithmUsagePolicy):
    def __init__(self, policy):
        self.policy = policy

    def digest_combination_allowed(
        self,
        signature_algo: algos.SignedDigestAlgorithm,
        message_digest_algo: algos.DigestAlgorithm,
        moment: Optional[datetime],
    ) -> AlgorithmUsageConstraint:
        return AlgorithmUsageConstraint(allowed=True)

    def digest_algorithm_allowed(
        self, algo: algos.DigestAlgorithm, moment: Optional[datetime]
    ) -> AlgorithmUsageConstraint:
        return self.policy.digest_algorithm_allowed(algo, moment)

    def signature_algorithm_allowed(
        self,
        algo: algos.SignedDigestAlgorithm,
        moment: Optional[datetime],
        public_key: Optional[keys.PublicKeyInfo],
    ) -> AlgorithmUsageConstraint:
        return self.policy.signature_algorithm_allowed(algo, moment, public_key)


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_digest_mismatch_allowed_but_inner_banned():
    output = await _generate_sig_with_mismatching_digest_algos()
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]

    policy = DisallowWeakAlgorithmsPolicy(weak_hash_algos=frozenset(['md5']))

    with pytest.raises(SignatureValidationError, match="md5.*not allowed"):
        await async_validate_pdf_signature(
            emb,
            SIMPLE_V_CONTEXT(),
            skip_diff=True,
            algorithm_policy=_AllowMismatches(policy),
        )


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_digest_mismatch_allowed_but_outer_banned():
    output = await _generate_sig_with_mismatching_digest_algos()
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]

    policy = DisallowWeakAlgorithmsPolicy(
        # yes, yes, I know. It's a test.
        weak_hash_algos=frozenset(['sha256'])
    )

    with pytest.raises(SignatureValidationError, match="sha256.*not allowed"):
        await async_validate_pdf_signature(
            emb,
            SIMPLE_V_CONTEXT(),
            skip_diff=True,
            algorithm_policy=_AllowMismatches(policy),
        )


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_digest_mismatch_allowed_passed_through_context():
    output = await _generate_sig_with_mismatching_digest_algos()
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]

    policy = DisallowWeakAlgorithmsPolicy(weak_hash_algos=frozenset())

    lenient_vc = ValidationContext(
        trust_roots=[ROOT_CERT], algorithm_usage_policy=_AllowMismatches(policy)
    )

    result = await async_validate_pdf_signature(
        emb,
        lenient_vc,
        skip_diff=True,
    )
    assert result.valid and result.intact


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_weak_sig_digest():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1', md_algorithm='md5')
    out = await signers.async_sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    with pytest.raises(DisallowedAlgorithmError):
        await async_val_trusted(emb)


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_weak_sig_digest_allowed():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1', md_algorithm='md5')
    out = await signers.async_sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]

    lenient_vc = ValidationContext(
        trust_roots=[ROOT_CERT], weak_hash_algos=set()
    )
    status = await async_val_trusted(emb, vc=lenient_vc)
    assert status.trusted


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_weak_sig_digest_disallowed_by_custom_policy():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1', md_algorithm='md5')
    out = await signers.async_sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]

    class Policy(CMSAlgorithmUsagePolicy):
        def digest_algorithm_allowed(
            self, algo: algos.DigestAlgorithm, moment: Optional[datetime]
        ) -> AlgorithmUsageConstraint:
            return AlgorithmUsageConstraint(
                allowed=False, failure_reason="Test reason"
            )

        def signature_algorithm_allowed(
            self,
            algo: algos.SignedDigestAlgorithm,
            moment: Optional[datetime],
            public_key: Optional[keys.PublicKeyInfo],
        ) -> AlgorithmUsageConstraint:
            return AlgorithmUsageConstraint(allowed=True)

    lenient_vc = ValidationContext(
        trust_roots=[ROOT_CERT], algorithm_usage_policy=Policy()
    )
    with pytest.raises(DisallowedAlgorithmError, match="Test reason"):
        await async_val_trusted(emb, vc=lenient_vc)


@pytest.mark.parametrize("with_issser", [False, True])
@freeze_time('2020-11-01')
def test_old_style_signing_cert_attr_ok(with_issser):
    if with_issser:
        fname = 'pades-with-old-style-signing-cert-attr-issser.pdf'
    else:
        # this file has an old-style signing cert attr without issuerSerial
        fname = 'pades-with-old-style-signing-cert-attr.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as f:
        r = PdfFileReader(f)
        s = r.embedded_signatures[0]
        assert s.field_name == 'Sig1'
        val_trusted(s)


@pytest.mark.parametrize("with_issser", [False, True])
@freeze_time('2020-11-01')
def test_old_style_signing_cert_attr_mismatch(with_issser):
    if with_issser:
        # this file has an old-style signing cert attr with issuerSerial
        fname = 'pades-with-old-style-signing-cert-attr-issser.pdf'
    else:
        fname = 'pades-with-old-style-signing-cert-attr.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as f:
        r = PdfFileReader(f)
        s = r.embedded_signatures[0]
        signer_info = s.signer_info
        digest = s.compute_digest()
    # signer1-long has the same key as signer1
    alt_cert = TESTING_CA.get_cert(CertLabel('signer1-long'))
    signer_info['sid'] = {
        'issuer_and_serial_number': cms.IssuerAndSerialNumber(
            {'issuer': alt_cert.issuer, 'serial_number': alt_cert.serial_number}
        )
    }
    with pytest.raises(
        SignatureValidationError,
        match="Signing certificate attribute does not match ",
    ) as exc_info:
        validate_sig_integrity(
            signer_info,
            alt_cert,
            expected_content_type='data',
            actual_digest=digest,
        )

    assert exc_info.value.ades_status == AdESStatus.INDETERMINATE
    assert (
        exc_info.value.ades_subindication
        == AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
    )


def test_old_style_signing_cert_attr_get():
    cert = TESTING_CA.get_cert(CertLabel('signer1'))
    v2 = as_signing_certificate_v2(cert, 'sha1')['certs'][0]
    v1 = as_signing_certificate(cert)['certs'][0]
    assert v1['issuer_serial'].dump() == v2['issuer_serial'].dump()
    assert v1['cert_hash'].dump() == v2['cert_hash'].dump()


def test_signing_cert_attr_malformed_issuer():
    from asn1crypto import x509

    cert = TESTING_CA.get_cert(CertLabel('signer1'))
    bogus_attr = as_signing_certificate_v2(cert)
    bogus_attr['certs'][0]['issuer_serial']['issuer'][0] = x509.GeneralName(
        {'dns_name': 'www.example.com'}
    )
    output = _tamper_with_signed_attrs(
        'signing_certificate_v2', resign=True, replace_with=bogus_attr
    )
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()
    with pytest.raises(
        SignatureValidationError,
        match="Signing certificate attribute does not match ",
    ) as exc_info:
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)
    assert exc_info.value.ades_status == AdESStatus.INDETERMINATE
    assert (
        exc_info.value.ades_subindication
        == AdESIndeterminate.NO_SIGNING_CERTIFICATE_FOUND
    )


def test_signing_cert_attr_duplicated():
    output = _tamper_with_signed_attrs(
        'signing_certificate_v2', resign=True, duplicate=True
    )
    r = PdfFileReader(output)
    emb = r.embedded_signatures[0]
    digest = emb.compute_digest()
    with pytest.raises(
        SignatureValidationError, match="Wrong cardinality for signing cert"
    ):
        validate_sig_integrity(emb.signer_info, emb.signer_cert, 'data', digest)


def test_verify_sig_without_signed_attrs():
    # pyHanko never produces signatures of this type, but we should be able
    # to validate them (this file was created using a modified version of
    # pyHanko's signing code, which will never see the light of day)

    with open(PDF_DATA_DIR + '/sig-no-signed-attrs.pdf', 'rb') as f:
        r = PdfFileReader(f)
        s = r.embedded_signatures[0]
        assert s.field_name == 'Sig1'
        val_untrusted(s)


def test_verify_sig_with_ski_sid():
    with open(PDF_DATA_DIR + '/sig-with-ski-sid.pdf', 'rb') as f:
        r = PdfFileReader(f)
        s = r.embedded_signatures[0]
        assert s.field_name == 'Sig1'
        val_untrusted(s)


@freeze_time('2020-11-01')
def test_sign_with_ecdsa_trust():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=FROM_ECC_CA
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    si = s.signer_info
    assert si['signature_algorithm']['algorithm'].native == 'sha384_ecdsa'
    val_trusted(s, vc=SIMPLE_ECC_V_CONTEXT())


@freeze_time('2020-11-01')
def test_sign_with_dsa_trust():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=FROM_DSA_CA
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    si = s.signer_info
    assert si['signature_algorithm']['algorithm'].native == 'sha256_dsa'
    val_trusted(s, vc=SIMPLE_DSA_V_CONTEXT())


@freeze_time('2020-11-01')
def test_sign_with_explicit_ecdsa_implied_hash():
    signer = signers.SimpleSigner(
        signing_cert=TESTING_CA_ECDSA.get_cert(CertLabel('signer1')),
        signing_key=TESTING_CA_ECDSA.key_set.get_private_key(
            KeyLabel('signer1')
        ),
        cert_registry=SimpleCertificateStore.from_certs(
            [ECC_ROOT_CERT, ECC_INTERM_CERT]
        ),
        # this is not allowed, but the validator should accept it anyway
        signature_mechanism=SignedDigestAlgorithm({'algorithm': 'ecdsa'}),
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=signer
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    si = s.signer_info
    assert si['signature_algorithm']['algorithm'].native == 'ecdsa'
    assert si['digest_algorithm']['algorithm'].native == 'sha384'
    assert s.field_name == 'Sig1'
    val_trusted(s, vc=SIMPLE_ECC_V_CONTEXT())


@freeze_time('2020-11-01')
def test_sign_with_explicit_dsa_implied_hash():
    signer = signers.SimpleSigner(
        signing_cert=TESTING_CA_DSA.get_cert(CertLabel('signer1')),
        signing_key=TESTING_CA_DSA.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [DSA_ROOT_CERT, DSA_INTERM_CERT]
        ),
        # this is not allowed, but the validator should accept it anyway
        signature_mechanism=SignedDigestAlgorithm({'algorithm': 'dsa'}),
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=signer
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    si = s.signer_info
    assert si['signature_algorithm']['algorithm'].native == 'dsa'
    assert s.field_name == 'Sig1'
    val_trusted(s, vc=SIMPLE_DSA_V_CONTEXT())


def test_sign_pss():
    signer = signers.SimpleSigner.load(
        CRYPTO_DATA_DIR + '/selfsigned.key.pem',
        CRYPTO_DATA_DIR + '/selfsigned.cert.pem',
        ca_chain_files=(CRYPTO_DATA_DIR + '/selfsigned.cert.pem',),
        key_passphrase=b'secret',
        prefer_pss=True,
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    sda: SignedDigestAlgorithm = emb.signer_info['signature_algorithm']
    assert sda.signature_algo == 'rsassa_pss'
    val_untrusted(emb)


def test_sign_pss_md_discrepancy():
    # Acrobat refuses to validate PSS signatures where the internal
    # hash functions disagree, but mathematically speaking, that shouldn't
    # be an issue.
    signer = signers.SimpleSigner.load(
        CRYPTO_DATA_DIR + '/selfsigned.key.pem',
        CRYPTO_DATA_DIR + '/selfsigned.cert.pem',
        ca_chain_files=(CRYPTO_DATA_DIR + '/selfsigned.cert.pem',),
        key_passphrase=b'secret',
        signature_mechanism=SignedDigestAlgorithm(
            {
                'algorithm': 'rsassa_pss',
                'parameters': RSASSAPSSParams(
                    {
                        'mask_gen_algorithm': MaskGenAlgorithm(
                            {
                                'algorithm': 'mgf1',
                                'parameters': DigestAlgorithm(
                                    {'algorithm': 'sha512'}
                                ),
                            }
                        ),
                        'hash_algorithm': DigestAlgorithm(
                            {'algorithm': 'sha256'}
                        ),
                        'salt_length': 478,
                    }
                ),
            }
        ),
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    sda: SignedDigestAlgorithm = emb.signer_info['signature_algorithm']
    assert sda.signature_algo == 'rsassa_pss'
    val_untrusted(emb)


@freeze_time('2020-11-01')
def test_direct_pdfcmsembedder_usage():
    # CMS-agnostic signing example
    #
    # write an in-place certification signature using the PdfCMSEmbedder
    # low-level API directly.

    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)

    # Phase 1: coroutine sets up the form field
    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    sig_field_ref = next(cms_writer)

    # just for kicks, let's check
    assert sig_field_ref.get_object()['/T'] == 'Signature'

    # Phase 2: make a placeholder signature object,
    # wrap it up together with the MDP config we want, and send that
    # on to cms_writer
    timestamp = datetime.now(tz=tzlocal.get_localzone())
    sig_obj = signers.SignatureObject(timestamp=timestamp, bytes_reserved=8192)

    md_algorithm = 'sha256'
    cms_writer.send(
        cms_embedder.SigObjSetup(
            sig_placeholder=sig_obj,
            mdp_setup=cms_embedder.SigMDPSetup(
                md_algorithm=md_algorithm,
                certify=True,
                docmdp_perms=fields.MDPPerm.NO_CHANGES,
            ),
        )
    )

    # Phase 3: write & hash the document (with placeholder)
    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm, in_place=True)
    )

    # Phase 4: construct CMS signature object, and pass it on to cms_writer

    # NOTE: I'm using a regular SimpleSigner here, but you can substitute
    # whatever CMS supplier you want.

    signer: signers.SimpleSigner = FROM_CA
    # let's supply the CMS object as a raw bytestring
    with pytest.deprecated_call():
        # noinspection PyDeprecation
        cms_bytes = signer.sign(
            data_digest=prep_digest.document_digest,
            digest_algorithm=md_algorithm,
            timestamp=timestamp,
        ).dump()
    sig_contents = cms_writer.send(cms_bytes)

    # we requested in-place output
    assert output is input_buf

    r = PdfFileReader(input_buf)
    val_trusted(r.embedded_signatures[0])

    # add some stuff to the DSS for kicks
    DocumentSecurityStore.add_dss(
        output, sig_contents, certs=FROM_CA.cert_registry, ocsps=(FIXED_OCSP,)
    )
    r = PdfFileReader(input_buf)
    dss = DocumentSecurityStore.read_dss(handler=r)
    val_trusted(r.embedded_signatures[0], extd=True)
    assert dss is not None
    assert len(dss.certs) == 3
    assert len(dss.ocsps) == 1


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_no_embed_root():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    cr = SimpleCertificateStore()
    cr.register_multiple(FROM_CA.cert_registry)
    no_embed_root_signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=cr,
        embed_roots=False,
    )
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1'),
        signer=no_embed_root_signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    assert '/AP' not in s.sig_field
    assert len(s.signed_data['certificates']) == 2
    await async_val_trusted(s)


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_noop_attribute_prov():
    class NoopProv(CMSAttributeProvider):
        async def build_attr_value(self, dry_run=False):
            return None

    class CustomSigner(signers.SimpleSigner):
        def _signed_attr_providers(self, *args, **kwargs):
            yield from super()._signed_attr_providers(*args, **kwargs)
            yield NoopProv()

    signer = CustomSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
    )
    input_data = b'Hello world!'
    signature = await signer.async_sign_general_data(input_data, 'sha256')

    # re-parse just to make sure we're starting fresh
    signature = cms.ContentInfo.load(signature.dump())

    raw_digest = hashlib.sha256(input_data).digest()
    content = signature['content']
    status = await async_validate_cms_signature(content, raw_digest=raw_digest)
    assert status.valid
    assert status.intact


@pytest.mark.parametrize('delete', [True, False])
@pytest.mark.asyncio
async def test_no_certificates(delete):
    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)
    md_algorithm = 'sha256'

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)
    sig_obj = signers.SignatureObject(bytes_reserved=8192)

    cms_writer.send(cms_embedder.SigObjSetup(sig_placeholder=sig_obj))

    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm, in_place=True)
    )

    signer: signers.SimpleSigner = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    cms_obj = await signer.async_sign(
        data_digest=prep_digest.document_digest,
        digest_algorithm=md_algorithm,
    )
    sd = cms_obj['content']
    if delete:
        del sd['certificates']
    else:
        sd['certificates'] = cms.CertificateSet([])
    cms_writer.send(cms_obj)

    r = PdfFileReader(output)
    with pytest.raises(CMSExtractionError, match='signer cert.*includ'):
        emb = r.embedded_signatures[0]
        await collect_validation_info(
            embedded_sig=emb, validation_context=ValidationContext()
        )
    with pytest.raises(SignatureValidationError, match='signer cert.*includ'):
        r.embedded_signatures[0].signer_cert.dump()


@pytest.mark.asyncio
async def test_two_signer_infos():
    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)
    md_algorithm = 'sha256'

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)
    sig_obj = signers.SignatureObject(bytes_reserved=8192)

    cms_writer.send(cms_embedder.SigObjSetup(sig_placeholder=sig_obj))

    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm, in_place=True)
    )

    signer: signers.SimpleSigner = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=SimpleCertificateStore(),
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    cms_obj = await signer.async_sign(
        data_digest=prep_digest.document_digest,
        digest_algorithm=md_algorithm,
    )
    sd = cms_obj['content']
    si = sd['signer_infos'][0]
    sd['signer_infos'] = [si, si]
    cms_writer.send(cms_obj)

    r = PdfFileReader(output)
    with pytest.raises(CMSExtractionError, match='exactly one'):
        emb = r.embedded_signatures[0]
        await collect_validation_info(
            embedded_sig=emb, validation_context=ValidationContext()
        )


def get_ac_aware_signer(actual_signer='signer1'):
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    signer = signers.SimpleSigner(
        signing_cert=pki_arch.get_cert(CertLabel(actual_signer)),
        signing_key=pki_arch.key_set.get_private_key(KeyLabel(actual_signer)),
        cert_registry=SimpleCertificateStore.from_certs(
            [
                pki_arch.get_cert('root'),
                pki_arch.get_cert('interm'),
                pki_arch.get_cert('root-aa'),
                pki_arch.get_cert('interm-aa'),
                pki_arch.get_cert('leaf-aa'),
            ]
        ),
        attribute_certs=[
            pki_arch.get_attr_cert(CertLabel('alice-role-with-rev'))
        ],
    )
    return signer


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_embed_ac(requests_mock):
    signer = get_ac_aware_signer()
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=signer
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.signed_data['version'].native == 'v4'
    # 4 CA certs, 1 AA certs, 1 AC, 1 signer cert -> 7 certs
    assert len(s.other_embedded_certs) == 5  # signer cert is excluded
    assert len(s.embedded_attr_certs) == 1
    main_vc, ac_vc = live_ac_vcs(requests_mock)
    status = await async_validate_pdf_signature(
        s, signer_validation_context=main_vc, ac_validation_context=ac_vc
    )
    assert status.bottom_line
    roles = list(status.ac_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert role['role_name'].native == 'bigboss@example.com'


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_embed_ac_revinfo_adobe_style(requests_mock):
    signer = get_ac_aware_signer()
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pki_arch = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-with-aa'))
    dummy_ts = timestamps.DummyTimeStamper(
        tsa_cert=pki_arch.get_cert(CertLabel('tsa')),
        tsa_key=pki_arch.key_set.get_private_key(KeyLabel('tsa')),
        certs_to_embed=SimpleCertificateStore.from_certs(
            [pki_arch.get_cert('root')]
        ),
    )
    from certomancer.integrations.illusionist import Illusionist

    from pyhanko_certvalidator.fetchers.requests_fetchers import (
        RequestsFetcherBackend,
    )

    fetchers = RequestsFetcherBackend().get_fetchers()
    main_vc = ValidationContext(
        trust_roots=[pki_arch.get_cert('root')],
        allow_fetching=True,
        other_certs=signer.cert_registry,
        fetchers=fetchers,
        revocation_mode='require',
    )
    ac_vc = ValidationContext(
        trust_roots=[pki_arch.get_cert('root-aa')],
        allow_fetching=True,
        other_certs=signer.cert_registry,
        fetchers=fetchers,
        revocation_mode='require',
    )
    Illusionist(pki_arch).register(requests_mock)
    out = await signers.async_sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            embed_validation_info=True,
            validation_context=main_vc,
            ac_validation_context=ac_vc,
        ),
        timestamper=dummy_ts,
        signer=signer,
    )

    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    # 4 CA certs, 1 AA certs, 1 AC, 1 signer cert -> 7 certs
    assert len(s.other_embedded_certs) == 5  # signer cert is excluded
    assert len(s.embedded_attr_certs) == 1
    from pyhanko.sign.validation import RevocationInfoValidationType

    status = await async_validate_pdf_ltv_signature(
        s,
        RevocationInfoValidationType.ADOBE_STYLE,
        validation_context_kwargs={'trust_roots': [pki_arch.get_cert('root')]},
        ac_validation_context_kwargs={
            'trust_roots': [pki_arch.get_cert('root-aa')]
        },
    )
    assert status.bottom_line
    roles = list(status.ac_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert role['role_name'].native == 'bigboss@example.com'


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_ac_detached(requests_mock):
    input_data = b'Hello world!'
    signer = get_ac_aware_signer()
    output = await signer.async_sign_general_data(input_data, 'sha256')
    assert output['content']['version'].native == 'v4'
    main_vc, ac_vc = live_ac_vcs(requests_mock)
    status = await async_validate_detached_cms(
        input_data,
        output['content'],
        signer_validation_context=main_vc,
        ac_validation_context=ac_vc,
    )
    assert status.bottom_line
    roles = list(status.ac_attrs['role'].attr_values)
    role = roles[0]
    assert isinstance(role, cms.RoleSyntax)
    assert len(list(status.ac_attrs)) == 1
    assert role['role_name'].native == 'bigboss@example.com'


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_ac_attr_validation_fail(requests_mock):
    input_data = b'Hello world!'
    signer = get_ac_aware_signer()
    output = await signer.async_sign_general_data(input_data, 'sha256')
    main_vc, ac_vc = live_ac_vcs(requests_mock)
    status = await async_validate_detached_cms(
        input_data,
        output['content'],
        signer_validation_context=main_vc,
        ac_validation_context=main_vc,  # pass in the wrong VC on purpose
    )
    assert status.bottom_line  # this should still be OK
    assert len(list(status.ac_attrs)) == 0
    assert 'role' not in status.ac_attrs  # ...but the attribute check fails


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_ac_attr_validation_holder_mismatch(requests_mock):
    input_data = b'Hello world!'
    # sign with a key pair that's not the same as the holder of the AC
    # that we're embedding
    signer = get_ac_aware_signer('signer2')
    output = await signer.async_sign_general_data(input_data, 'sha256')
    main_vc, ac_vc = live_ac_vcs(requests_mock)
    status = await async_validate_detached_cms(
        input_data,
        output['content'],
        signer_validation_context=main_vc,
        ac_validation_context=ac_vc,
    )
    assert status.bottom_line  # this should still be OK
    assert len(list(status.ac_attrs)) == 0
    assert 'role' not in status.ac_attrs  # ...but the attribute check fails


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cades_cms_with_tst():
    signature = await FROM_CA.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
        timestamper=DUMMY_TS,
        use_cades=True,
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!', signature['content']
    )
    assert status.signer_reported_dt == datetime.now(tz=timezone.utc)
    assert status.timestamp_validity.intact
    assert status.timestamp_validity.valid
    assert status.timestamp_validity.timestamp == datetime.now(tz=timezone.utc)
    assert 'The TSA certificate is untrusted' in status.pretty_print_details()
    assert status.valid
    assert status.intact


class GenericOpenTypePair(core.Sequence):
    _fields = [('f1', core.Any), ('f2', core.Any)]


NONSENSICAL_ATTR = cms.AttCertAttribute.load(
    GenericOpenTypePair(
        {
            'f1': core.ObjectIdentifier('2.5.4.72'),  # 'role'
            'f2': cms.SetOfOctetString(
                [core.OctetString(b'This makes no sense')]
            ),
        }
    ).dump()
)

UNTYPABLE_ATTR = cms.AttCertAttribute.load(
    GenericOpenTypePair(
        {
            'f1': core.OctetString(b'This makes even less sense'),
            'f2': cms.SetOfOctetString(
                [core.OctetString(b'This makes no sense')]
            ),
        }
    ).dump()
)


@pytest.mark.parametrize(
    'data,fatal',
    list(
        itertools.product(
            [
                (NONSENSICAL_ATTR, 'Failed to parse claimed.*role'),
                (UNTYPABLE_ATTR, 'Failed to parse.*unknown type'),
            ],
            [True, False],
        )
    ),
)
@pytest.mark.asyncio
async def test_parse_malformed_claimed_attrs(data, fatal):
    # This should parse up to the first level and be reencoded by asn1crypto
    #  without asking any questions.
    nonsensical_attr, msg = data
    claimed_attrs = [nonsensical_attr, SAMPLE_GROUP_ATTR]
    if fatal:
        with pytest.raises(SignatureValidationError, match=msg):
            ClaimedAttributes.from_iterable(
                claimed_attrs, parse_error_fatal=True
            )
    else:
        claimed = ClaimedAttributes.from_iterable(
            claimed_attrs, parse_error_fatal=False
        )
        # The malformed attribute shouldn't have been processed,
        # but the other attrs should've
        assert len(claimed) == 1


@pytest.mark.parametrize('bad_attr', [NONSENSICAL_ATTR, UNTYPABLE_ATTR])
@pytest.mark.asyncio
async def test_validate_with_malformed_claimed_attrs(bad_attr, requests_mock):
    # This should parse up to the first level and be reencoded by asn1crypto
    #  without asking any questions.
    cms_sig = await FROM_CA.async_sign_general_data(
        b'Hello world',
        digest_algorithm='sha256',
        signed_attr_settings=PdfCMSSignedAttributes(
            cades_signed_attrs=CAdESSignedAttrSpec(
                signer_attributes=SignerAttrSpec(
                    claimed_attrs=[bad_attr, SAMPLE_GROUP_ATTR],
                    certified_attrs=[],
                )
            )
        ),
    )
    status = await async_validate_detached_cms(
        input_data=b'Hello world',
        signed_data=cms_sig['content'],
        signer_validation_context=live_testing_vc(requests_mock),
    )
    assert isinstance(status, StandardCMSSignatureStatus)
    # The malformed attribute shouldn't have been processed,
    # but the other attrs should've
    assert len(status.cades_signer_attrs.claimed_attrs) == 1


BASIC_AC_ISSUER_SETUP = '''
  ac-issuer:
    subject: root
    issuer: root
    validity:
      valid-from: "2000-01-01T00:00:00+0000"
      valid-to: "2500-01-01T00:00:00+0000"
    extensions:
      - id: key_usage
        critical: true
        smart-value:
          schema: key-usage
          params: [digital_signature]
  signer1:
    subject: signer1
    issuer: root
    validity:
      valid-from: "2000-01-01T00:00:00+0000"
      valid-to: "2100-01-01T00:00:00+0000"
    extensions:
      - id: key_usage
        critical: true
        smart-value:
          schema: key-usage
          params: [digital_signature]
'''


@pytest.mark.asyncio
async def test_parse_ac_with_malformed_attribute(requests_mock):
    attr_cert_cfg = f'''
    test-ac:
      holder:
          name: signer1
          # this needs to match against something from a totally different PKI
          # arch, so make the coupling as loose as possible
          include-base-cert-id: false
          include-entity-name: true
      issuer: root
      attributes:
          - id: charging_identity
            smart-value:
                schema: ietf-attribute
                params: ["Big Corp Inc."]
      validity:
        valid-from: "2000-01-01T00:00:00+0000"
        valid-to: "2100-01-01T00:00:00+0000"
    '''

    pki_arch = PKIArchitecture(
        arch_label=ArchLabel('test'),
        key_set=TESTING_CA.key_set,
        entities=TESTING_CA.entities,
        cert_spec_config=yaml.safe_load(BASIC_AC_ISSUER_SETUP),
        ac_spec_config=yaml.safe_load(attr_cert_cfg),
        service_config={},
        external_url_prefix='http://test.test',
    )
    spec = pki_arch.get_attr_cert_spec(CertLabel('test-ac'))

    # we have to get a bit creative to get Certomancer to output invalid asn1
    #  in exactly the way we want
    class FakeAttrSpec:
        # noinspection PyUnusedLocal
        def to_asn1(self, arch):
            return NONSENSICAL_ATTR

    # noinspection PyTypeChecker
    spec.attributes.append(FakeAttrSpec())

    cms_sig = await FROM_CA.async_sign_general_data(
        b'Hello world',
        digest_algorithm='sha256',
        signed_attr_settings=PdfCMSSignedAttributes(
            cades_signed_attrs=CAdESSignedAttrSpec(
                signer_attributes=SignerAttrSpec(
                    certified_attrs=[
                        pki_arch.get_attr_cert(CertLabel('test-ac'))
                    ],
                    claimed_attrs=[],
                )
            )
        ),
    )
    vc = live_testing_vc(requests_mock)
    ac_vc = ValidationContext(
        trust_roots=[pki_arch.get_cert(CertLabel('ac-issuer'))],
        allow_fetching=False,
    )
    status = await async_validate_detached_cms(
        input_data=b'Hello world',
        signed_data=cms_sig['content'],
        signer_validation_context=vc,
        ac_validation_context=ac_vc,
    )
    assert isinstance(status, StandardCMSSignatureStatus)
    # The malformed attribute shouldn't have been processed,
    # but the other attrs should've
    assert len(status.cades_signer_attrs.certified_attrs) == 1


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_invalid_cn():
    cert = TESTING_CA_ERRORS.get_cert(CertLabel('signer-with-invalid-cn'))
    sgn = signers.SimpleSigner(
        signing_cert=cert,
        signing_key=TESTING_CA.key_set.get_private_key(KeyLabel('signer2')),
        cert_registry=SimpleCertificateStore.from_certs(
            [ROOT_CERT, INTERM_CERT]
        ),
    )
    signature = await sgn.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
    )
    signature = cms.ContentInfo.load(signature.dump())
    status = await async_validate_detached_cms(
        b'Hello world!',
        signature['content'],
        signer_validation_context=SIMPLE_V_CONTEXT(),
    )
    assert status.valid
    assert status.intact
    # currently, pyhanko-certvalidator does not handle invalid cases like this
    # gracefully.
    assert not status.trusted


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_invalid_cn_in_ca_wrong_cert():
    cert = TESTING_CA_ERRORS.get_cert(CertLabel('issued-by-invalid-cn'))
    sgn = signers.SimpleSigner(
        signing_cert=cert,
        signing_key=TESTING_CA.key_set.get_private_key(KeyLabel('signer2')),
        cert_registry=SimpleCertificateStore.from_certs(
            # this is the wrong intermediate cert, so it shouldn't match
            # (the actual CA cert can't be added to a certificate store
            #  because its subject isn't hashable)
            [ROOT_CERT, INTERM_CERT]
        ),
    )
    signature = await sgn.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=False,
    )
    signature = cms.ContentInfo.load(signature.dump())
    with pytest.raises(
        SignatureValidationError,
        match="Signing certificate attribute does not match",
    ):
        await async_validate_detached_cms(
            b'Hello world!',
            signature['content'],
            signer_validation_context=SIMPLE_V_CONTEXT(),
        )


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_detached_cms_with_invalid_cn_in_ca():
    cert = TESTING_CA_ERRORS.get_cert(CertLabel('issued-by-invalid-cn'))

    class DummySigner(signers.SimpleSigner):
        # make sure the signing cert attr is not included
        def _signed_attr_providers(self, *args, **kwargs):
            return ()

    sgn = DummySigner(
        signing_cert=cert,
        signing_key=TESTING_CA.key_set.get_private_key(KeyLabel('signer2')),
        cert_registry=SimpleCertificateStore.from_certs(
            # the malformed intermediate CA can't be added to a certificate
            # store, we'll do that later
            [ROOT_CERT]
        ),
    )
    signature = await sgn.async_sign_general_data(
        b'Hello world!',
        'sha256',
        detached=True,
    )

    signature['content']['certificates'].append(
        cms.CertificateChoices(
            name='certificate',
            value=TESTING_CA_ERRORS.get_cert(CertLabel('ca-with-invalid-cn')),
        )
    )
    signature = cms.ContentInfo.load(signature.dump())

    status = await async_validate_detached_cms(
        b'Hello world!',
        signature['content'],
        signer_validation_context=SIMPLE_V_CONTEXT(),
    )
    assert status.valid
    assert status.intact
    # Because the X.509 validation chokes on those intermediate certs
    #  (already very early, when constructing the certificate registry)
    # But that's something I'm willing to live with not supporting.
    assert not status.trusted


@pytest.mark.parametrize(
    'testing_ca, expected_md',
    [
        (TESTING_CA, 'sha256'),
        (TESTING_CA_ECDSA, 'sha384'),
        (TESTING_CA_ED25519, 'sha512'),
        (TESTING_CA_ED448, 'shake256'),
    ],
)
def test_key_based_digest_selection(testing_ca, expected_md):
    pubkey = testing_ca.key_set.get_public_key(KeyLabel('signer1'))
    md = select_suitable_signing_md(pubkey)
    assert md == expected_md


async def _generate_badly_ordered_signed_attrs(
    digest: bytes, signer: signers.Signer
):
    class HackySeq(core.SequenceOf):
        tag = 17  # SET OF
        _child_spec = cms.CMSAttribute

    class HackySignerInfo(core.Sequence):
        _fields = [
            ('version', cms.CMSVersion),
            ('sid', cms.SignerIdentifier),
            ('digest_algorithm', DigestAlgorithm),
            ('signed_attrs', core.Asn1Value, {'implicit': 0}),
            ('signature_algorithm', SignedDigestAlgorithm),
            ('signature', core.OctetString),
        ]

    signing_cert = signer.signing_cert
    seq = HackySeq(
        [
            simple_cms_attribute('content_type', cms.ContentType('data')),
            simple_cms_attribute('message_digest', core.OctetString(digest)),
            await attributes.SigningCertificateV2Provider(
                signing_cert=signing_cert
            ).get_attribute(),
            # lexicographically, this one can't be last if we would encode this
            # as a SET OF
            simple_cms_attribute(
                '2.999', core.OctetString(b"\xde\xad\xbe\xef")
            ),
        ]
    )

    signature = await signer.async_sign_raw(
        seq.dump(), digest_algorithm='sha256'
    )

    digest_algo_obj = cms.DigestAlgorithm({'algorithm': 'sha256'})
    algo = signer.get_signature_mechanism_for_digest('sha256')
    sig_info = HackySignerInfo(
        {
            'version': 'v1',
            'sid': cms.SignerIdentifier(
                {
                    'issuer_and_serial_number': cms.IssuerAndSerialNumber(
                        {
                            'issuer': signing_cert.issuer,
                            'serial_number': signing_cert.serial_number,
                        }
                    )
                }
            ),
            'digest_algorithm': digest_algo_obj,
            'signed_attrs': seq,
            'signature_algorithm': algo,
            'signature': signature,
        }
    )

    nonder_data = sig_info.dump()
    sig_info = cms.SignerInfo.load(nonder_data)
    return sig_info


@pytest.mark.asyncio
async def test_tolerate_der_deviations():
    msg = b'Hello world'
    digest = hashlib.sha256(msg).digest()
    sig_info = await _generate_badly_ordered_signed_attrs(digest, FROM_CA)

    intact, valid = validate_sig_integrity(
        signer_info=sig_info,
        cert=FROM_CA.signing_cert,
        expected_content_type='data',
        actual_digest=digest,
    )

    assert intact and valid


@pytest.mark.asyncio
@freeze_time('2020-11-01')
async def test_tolerate_der_deviations_in_pdf():
    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)
    md_algorithm = 'sha256'

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)
    sig_obj = signers.SignatureObject(bytes_reserved=8192)

    cms_writer.send(cms_embedder.SigObjSetup(sig_placeholder=sig_obj))

    prep_digest, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm, in_place=True)
    )

    signer: signers.SimpleSigner = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    cms_obj = await signer.async_sign(
        data_digest=prep_digest.document_digest,
        digest_algorithm=md_algorithm,
    )
    sd = cms_obj['content']
    sd['signer_infos'][0] = await _generate_badly_ordered_signed_attrs(
        digest=prep_digest.document_digest, signer=signer
    )
    cms_writer.send(cms_obj)

    # first, check if the test file is "corrupted" in the right way
    r = PdfFileReader(output)
    sd = r.embedded_signatures[0].signed_data
    raw_enc = sd.dump()
    forced_enc = sd.dump(force=True)
    # the 2.999: DEADBEEF attribute should be early in the DER version, but
    # appear only much further in the non-DER version
    marker = b'\x31\x06\x04\x04\xde\xad\xbe\xef'
    assert raw_enc.find(marker) > forced_enc.find(marker)

    # now we run the actual validation after reopening the file
    r = PdfFileReader(output)
    await async_val_trusted(r.embedded_signatures[0])


@freeze_time('2020-11-01')
def test_ed448_no_length():
    # verify that we still accept ed448 with id-shake256 without paramaters

    fname = os.path.join(PDF_DATA_DIR, 'ed448-shake256-nolen.pdf')
    with open(fname, 'rb') as inf:
        r = PdfFileReader(inf)
        s = r.embedded_signatures[0]
        status = val_untrusted(s)
    assert status.md_algorithm == 'shake256'

    assert len(s.external_digest) == 64


@freeze_time('2020-11-01')
def test_ed448_invalid_hash_algo_validation():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))

    with pytest.raises(
        DisallowedAlgorithmError, match='algorithm.*does not match'
    ):
        fname = os.path.join(PDF_DATA_DIR, 'ed448-disallowed-hash.pdf')
        with open(fname, 'rb') as inf:
            r = PdfFileReader(inf)
            s = r.embedded_signatures[0]
            val_untrusted(s)



from asn1crypto.algos import KdfAlgorithmId
from asn1crypto.cms import (
    CMSAttribute,
    CMSAttributeType,
    ContentInfo,
    ContentType,
    EncapsulatedContentInfo,
    SetOfContentInfo,
)
from asn1crypto.core import Integer, OctetString, Sequence


class PdfMacIntegrityInfo(Sequence):
    _fields = [
        ('version', Integer),
        ('data_digest', OctetString),
        ('signature_digest', OctetString, {'implicit': 0, 'optional': True}),
    ]


def asn1crypto_setup():
    # This is the content type used for the MAC'd data in a PDF MAC token
    ContentType._map['1.0.32004.1.0'] = 'pdf_mac_integrity_info'
    if ContentType._reverse_map is not None:
        ContentType._reverse_map['pdf_mac_integrity_info'] = (
            '1.0.32004.1.0'  # pragma: nocover
        )
    EncapsulatedContentInfo._oid_specs['pdf_mac_integrity_info'] = (
        ContentInfo._oid_specs['pdf_mac_integrity_info']
    ) = PdfMacIntegrityInfo

    # This OID represents the key derivation function defined in ISO 32004
    # to derive the MAC key encryption key from the PDF's file encryption key
    #  (+ other data)
    KdfAlgorithmId._map['1.0.32004.1.1'] = 'pdf_mac_wrap_kdf'
    if KdfAlgorithmId._reverse_map is not None:
        KdfAlgorithmId._reverse_map['pdf_mac_wrap_kdf'] = (
            '1.0.32004.1.1'  # pragma: nocover
        )

    # This OID represents the attribute used when attaching PDF MAC tokens
    # to signatures.
    CMSAttributeType._map['1.0.32004.1.2'] = 'pdf_mac_data'
    if CMSAttributeType._reverse_map is not None:
        CMSAttributeType._reverse_map['1.0.32004.1.2'] = (
            'pdf_mac_data'  # pragma: nocover
        )
    CMSAttribute._oid_specs['pdf_mac_data'] = SetOfContentInfo


asn1crypto_setup()



"""
.. versionadded:: 0.14.0

This module contains the XMP data model classes and namespace registry,
in addition to a simplified document metadata model used for automated
metadata management.
"""

import enum
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, Iterable, Iterator, List, Optional, Tuple, Union

from pyhanko import __version__
from pyhanko.pdf_utils.misc import StringWithLanguage

__all__ = [
    'DocumentMetadata',
    'VENDOR',
    'MetaString',
    'ExpandedName',
    'Qualifiers',
    'XmpValue',
    'XmpStructure',
    'XmpArrayType',
    'XmpArray',
    'NS',
    'XML_LANG',
    'RDF_RDF',
    'RDF_SEQ',
    'RDF_BAG',
    'RDF_ALT',
    'RDF_LI',
    'RDF_VALUE',
    'RDF_RESOURCE',
    'RDF_PARSE_TYPE',
    'RDF_ABOUT',
    'RDF_DESCRIPTION',
    'DC_TITLE',
    'DC_CREATOR',
    'DC_DESCRIPTION',
    'PDF_PRODUCER',
    'PDF_KEYWORDS',
    'X_XMPMETA',
    'X_XMPTK',
    'XMP_CREATORTOOL',
    'XMP_CREATEDATE',
    'XMP_MODDATE',
]

VENDOR = 'pyHanko ' + __version__
"""
pyHanko version identifier in textual form
"""

MetaString = Union[StringWithLanguage, str, None]
"""
A regular string, a string with a language code, or nothing at all.
"""


@dataclass
class DocumentMetadata:
    """
    Simple representation of document metadata. All entries are optional.
    """

    title: MetaString = None
    """
    The document's title.
    """

    author: MetaString = None
    """
    The document's author.
    """

    subject: MetaString = None
    """
    The document's subject.
    """

    keywords: List[str] = field(default_factory=list)
    """
    Keywords associated with the document.
    """

    creator: MetaString = None
    """
    The software that was used to author the document.

    .. note::
        This is distinct from the producer, which is typically used to indicate
        which PDF processor(s) interacted with the file.
    """

    created: Union[str, datetime, None] = None
    """
    The time when the document was created. To set it to the current time,
    specify ``now``.
    """

    last_modified: Union[str, datetime, None] = "now"
    """
    The time when the document was last modified. Defaults to the current time
    upon serialisation if not specified.
    """

    xmp_extra: List['XmpStructure'] = field(default_factory=list)
    """
    Extra XMP metadata.
    """

    xmp_unmanaged: bool = False
    """
    Flag metadata as XMP-only. This means that the info dictionary will be
    cleared out as much as possible, and that all attributes other than
    :attr:`xmp_extra` will be ignored when updating XMP metadata.

    .. note::
        The last-modified date and producer entries
        in the info dictionary will still be updated.

    .. note::
        :class:`DocumentMetadata` represents a data model that is much more
        simple than what XMP is actually capable of. You can use this flag
        if you need more fine-grained control.
    """

    def view_over(self, base: 'DocumentMetadata'):
        return DocumentMetadata(
            title=self.title or base.title,
            author=self.author or base.author,
            subject=self.subject or base.subject,
            keywords=list(self.keywords or base.keywords),
            creator=self.creator or base.creator,
            created=self.created or base.created,
            last_modified=self.last_modified,
        )


@dataclass(frozen=True)
class ExpandedName:
    """
    An expanded XML name.
    """

    ns: str
    """
    The URI of the namespace in which the name resides.
    """

    local_name: str
    """
    The local part of the name.
    """

    def __str__(self):
        ns = self.ns
        sep = '' if ns.endswith('/') or ns.endswith('#') else '/'
        return f"{ns}{sep}{self.local_name}"

    def __repr__(self):
        return str(self)


NS = {
    'xml': 'http://www.w3.org/XML/1998/namespace',
    'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'xmp': 'http://ns.adobe.com/xap/1.0/',
    'dc': 'http://purl.org/dc/elements/1.1/',
    'pdf': 'http://ns.adobe.com/pdf/1.3/',
    'pdfaid': 'http://www.aiim.org/pdfa/ns/id/',
    'pdfuaid': 'http://www.aiim.org/pdfua/ns/id/',
    'pdfaSchema': 'http://www.aiim.org/pdfa/ns/schema#',
    'pdfaExtension': 'http://www.aiim.org/pdfa/ns/extension/',
    'pdfaProperty': 'http://www.aiim.org/pdfa/ns/property#',
    'x': 'adobe:ns:meta/',
}
"""
Known namespaces and their customary prefixes.
"""


XML_LANG = ExpandedName(ns=NS['xml'], local_name='lang')
"""
``lang`` in the ``xml`` namespace.
"""

RDF_RDF = ExpandedName(ns=NS['rdf'], local_name='RDF')
"""
``RDF`` in the ``rdf`` namespace.
"""

RDF_SEQ = ExpandedName(ns=NS['rdf'], local_name='Seq')
"""
``Seq`` in the ``rdf`` namespace.
"""

RDF_BAG = ExpandedName(ns=NS['rdf'], local_name='Bag')
"""
``Bag`` in the ``rdf`` namespace.
"""

RDF_ALT = ExpandedName(ns=NS['rdf'], local_name='Alt')
"""
``Alt`` in the ``rdf`` namespace.
"""

RDF_LI = ExpandedName(ns=NS['rdf'], local_name='li')
"""
``li`` in the ``rdf`` namespace.
"""

RDF_VALUE = ExpandedName(ns=NS['rdf'], local_name='value')
"""
``value`` in the ``rdf`` namespace.
"""

RDF_RESOURCE = ExpandedName(ns=NS['rdf'], local_name='resource')
"""
``resource`` in the ``rdf`` namespace.
"""

RDF_ABOUT = ExpandedName(ns=NS['rdf'], local_name='about')
"""
``about`` in the ``rdf`` namespace.
"""

RDF_PARSE_TYPE = ExpandedName(ns=NS['rdf'], local_name='parseType')
"""
``parseType`` in the ``rdf`` namespace.
"""

RDF_DESCRIPTION = ExpandedName(ns=NS['rdf'], local_name='Description')
"""
``Description`` in the ``rdf`` namespace.
"""

X_XMPMETA = ExpandedName(ns=NS['x'], local_name='xmpmeta')
"""
``xmpmeta`` in the ``x`` namespace.
"""

X_XMPTK = ExpandedName(ns=NS['x'], local_name='xmptk')
"""
``xmptk`` in the ``x`` namespace.
"""


DC_TITLE = ExpandedName(ns=NS['dc'], local_name='title')
"""
``title`` in the ``dc`` namespace.
"""

DC_CREATOR = ExpandedName(ns=NS['dc'], local_name='creator')
"""
``creator`` in the ``dc`` namespace.
"""

DC_DESCRIPTION = ExpandedName(ns=NS['dc'], local_name='description')
"""
``description`` in the ``dc`` namespace.
"""

PDF_KEYWORDS = ExpandedName(ns=NS['pdf'], local_name='keywords')
"""
``keywords`` in the ``pdf`` namespace.
"""

PDF_PRODUCER = ExpandedName(ns=NS['pdf'], local_name='Producer')
"""
``Producer`` in the ``pdf`` namespace.
"""

XMP_CREATORTOOL = ExpandedName(ns=NS['xmp'], local_name='CreatorTool')
"""
``CreatorTool`` in the ``xmp`` namespace.
"""

XMP_CREATEDATE = ExpandedName(ns=NS['xmp'], local_name='CreateDate')
"""
``CreateDate`` in the ``xmp`` namespace.
"""

XMP_MODDATE = ExpandedName(ns=NS['xmp'], local_name='ModifyDate')
"""
``ModifyDate`` in the ``xmp`` namespace.
"""


class Qualifiers:
    """
    XMP value qualifiers wrapper. Implements ``__getitem__``.
    Note that ``xml:lang`` gets special treatment.

    :param quals:
        The qualifiers to model.
    """

    _quals: Dict[ExpandedName, 'XmpValue']
    _lang: Optional[str]

    def __init__(self, quals: Dict[ExpandedName, 'XmpValue']):
        self._quals = quals
        try:
            lang = quals[XML_LANG]
            del quals[XML_LANG]
            if not isinstance(lang.value, str):
                raise TypeError  # pragma: nocover
            self._lang = lang.value
        except KeyError:
            self._lang = None

    @classmethod
    def of(cls, *lst: Tuple[ExpandedName, 'XmpValue']) -> 'Qualifiers':
        """
        Construct a :class:`.Qualifiers` object from a list of name-value pairs.

        :param lst:
            A list of name-value pairs.
        :return:
            A :class:`.Qualifiers` object.
        """
        return Qualifiers({k: v for k, v in lst})

    @classmethod
    def lang_as_qual(cls, lang: Optional[str]) -> 'Qualifiers':
        """
        Construct a :class:`.Qualifiers` object that only wraps a language
        qualifier.

        :param lang:
            A language code.
        :return:
            A :class:`.Qualifiers` object.
        """
        quals = Qualifiers({})
        if lang:
            quals._lang = lang
        return quals

    def __getitem__(self, item):
        return self._quals[item]

    def iter_quals(
        self, with_lang: bool = True
    ) -> Iterable[Tuple[ExpandedName, 'XmpValue']]:
        """
        Iterate over all qualifiers.

        :param with_lang:
            Include the language qualifier.
        :return:
        """
        yield from self._quals.items()
        if with_lang and self._lang is not None:
            yield XML_LANG, XmpValue(self._lang)

    @property
    def lang(self) -> Optional[str]:
        """
        Retrieve the language qualifier, if any.
        """
        return self._lang

    @property
    def has_non_lang_quals(self) -> bool:
        """
        Check if there are any non-language qualifiers.
        """
        return bool(self._quals)

    def __bool__(self):
        return bool(self._quals or self._lang)

    def __repr__(self):
        q = dict(self._quals)
        if self._lang:
            q['lang'] = self._lang
        return f"Qualifiers({q!r})"

    def __eq__(self, other):
        return (
            isinstance(other, Qualifiers)
            and self._lang == other._lang
            and self._quals == other._quals
        )


@dataclass(frozen=True)
class XmpUri:
    """
    An XMP URI value.
    """

    value: str

    def __str__(self):
        return self.value


@dataclass
class XmpValue:
    """
    A general XMP value, potentially with qualifiers.
    """

    value: Union['XmpStructure', 'XmpArray', XmpUri, str]
    """
    The value.
    """

    qualifiers: Qualifiers = field(default_factory=Qualifiers.of)
    """
    Qualifiers that apply to the value.
    """


class XmpStructure:
    """
    A generic XMP structure value. Implements ``__getitem__`` for field access.

    :param fields:
        The structure's fields.
    """

    # isomorphic to Qualifiers, but we keep them separate to stay
    # closer to the spec (and this one doesn't special-case anything)

    def __init__(self, fields: Dict[ExpandedName, 'XmpValue']):
        self._fields: Dict[ExpandedName, XmpValue] = fields

    @classmethod
    def of(cls, *lst: Tuple[ExpandedName, 'XmpValue']) -> 'XmpStructure':
        """
        Construct an :class:`.XmpStructure` from a list of name-value pairs.

        :param lst:
            A list of name-value pairs.
        :return:
            An an :class:`.XmpStructure`.
        """
        return cls({k: v for k, v in lst})

    def __getitem__(self, item):
        return self._fields[item]

    def __iter__(self) -> Iterator[Tuple[ExpandedName, 'XmpValue']]:
        yield from self._fields.items()

    def __repr__(self):
        return f"XmpStructure({self._fields!r})"

    def __eq__(self, other):
        return isinstance(other, XmpStructure) and self._fields == other._fields


@enum.unique
class XmpArrayType(enum.Enum):
    """
    XMP array types.
    """

    ORDERED = 'Seq'
    """
    Ordered array.
    """

    UNORDERED = 'Bag'
    """
    Unordered array.
    """

    ALTERNATIVE = 'Alt'
    """
    Alternative array.
    """

    def as_rdf(self) -> ExpandedName:
        """
        Render the type as an XML name.
        """
        return ExpandedName(ns=NS['rdf'], local_name=str(self.value))


@dataclass
class XmpArray:
    """
    An XMP array.
    """

    array_type: XmpArrayType
    """
    The type of the array.
    """

    entries: List[XmpValue]
    """
    The entries in the array.
    """

    @classmethod
    def ordered(cls, lst: Iterable[XmpValue]) -> 'XmpArray':
        """
        Convert a list to an ordered XMP array.

        :param lst:
            An iterable of XMP values.
        :return:
            An ordered :class:`.XmpArray`.
        """
        return cls(XmpArrayType.ORDERED, list(lst))

    @classmethod
    def unordered(cls, lst: Iterable[XmpValue]) -> 'XmpArray':
        """
        Convert a list to an unordered XMP array.

        :param lst:
            An iterable of XMP values.
        :return:
            An unordered :class:`.XmpArray`.
        """
        return cls(XmpArrayType.UNORDERED, list(lst))

    @classmethod
    def alternative(cls, lst: Iterable[XmpValue]) -> 'XmpArray':
        """
        Convert a list to an alternative XMP array.

        :param lst:
            An iterable of XMP values.
        :return:
            An alternative :class:`.XmpArray`.
        """
        return cls(XmpArrayType.ALTERNATIVE, list(lst))

    def __eq__(self, other):
        if (
            not isinstance(other, XmpArray)
            or self.array_type != other.array_type
        ):
            return False
        if self.array_type == XmpArrayType.UNORDERED:
            return all(e in self.entries for e in other.entries) and all(
                e in other.entries for e in self.entries
            )
        else:
            return self.entries == other.entries



"""
.. versionchanged:: 0.13.0
    Refactor ``crypt`` module into package.
.. versionchanged:: 0.3.0
    Added support for PDF 2.0 encryption standards and crypt filters.

Utilities for PDF encryption. This module covers all methods outlined in the
standard:

* Legacy RC4-based encryption (based on PyPDF2 code).
* AES-128 encryption with legacy key derivation (partly based on PyPDF2 code).
* PDF 2.0 AES-256 encryption.
* Public key encryption backed by any of the above.

Following the language in the standard, encryption operations are backed by
subclasses of the :class:`SecurityHandler` class, which provides a more or less
generic API.

.. danger::
    The members of this package are all considered internal API, and are
    therefore subject to change without notice.

.. danger::
    One should also be aware that the legacy encryption scheme implemented
    here is (very) weak, and we only support it for compatibility reasons.
    Under no circumstances should it still be used to encrypt new files.


About crypt filters
-------------------

Crypt filters are objects that handle encryption and decryption of streams and
strings, either for all of them, or for a specific subset (e.g. streams
representing embedded files). In the context of the PDF standard, crypt filters
are a notion that only makes sense for security handlers of version 4 and up.
In pyHanko, however, *all* encryption and decryption operations pass through
crypt filters, and the serialisation/deserialisation logic in
:class:`SecurityHandler` and its subclasses transparently deals with staying
backwards compatible with earlier revisions.

Internally, pyHanko loosely distinguishes between implicit and explicit
uses of crypt filters:

* Explicit crypt filters are used by directly referring to them from the
  ``/Filter`` entry of a stream dictionary. These are invoked in the usual
  stream decoding process.
* Implicit crypt filters are set by the ``/StmF`` and ``/StrF`` entries
  in the security handler's crypt filter configuration, and are invoked by the
  object reading/writing procedures as necessary. These filters are invisble
  to the stream encoding/decoding process: the
  :attr:`~.generic.StreamObject.encoded_data` attribute of
  an "implicitly encrypted" stream will therefore contain decrypted data ready
  to be decoded in the usual way.

As long as you don't require access to encoded object data and/or raw encrypted
object data, this distiction should be irrelevant to you as an API user.
"""

from .api import (
    IDENTITY,
    AuthResult,
    AuthStatus,
    CryptFilter,
    CryptFilterBuilder,
    CryptFilterConfiguration,
    IdentityCryptFilter,
    PdfKeyNotAvailableError,
    SecurityHandler,
    SecurityHandlerVersion,
    build_crypt_filter,
)
from .cred_ser import SerialisableCredential, SerialisedCredential
from .filter_mixins import AESCryptFilterMixin, RC4CryptFilterMixin
from .pubkey import (
    DEF_EMBEDDED_FILE,
    DEFAULT_CRYPT_FILTER,
    EnvelopeKeyDecrypter,
    PubKeyAdbeSubFilter,
    PubKeyAESCryptFilter,
    PubKeyCryptFilter,
    PubKeyRC4CryptFilter,
    PubKeySecurityHandler,
    SimpleEnvelopeKeyDecrypter,
)
from .standard import (
    STD_CF,
    StandardAESCryptFilter,
    StandardCryptFilter,
    StandardRC4CryptFilter,
    StandardSecurityHandler,
    StandardSecuritySettingsRevision,
    _PasswordCredential,
)

__all__ = [
    'SecurityHandler',
    'StandardSecurityHandler',
    'PubKeySecurityHandler',
    'AuthResult',
    'AuthStatus',
    'SecurityHandlerVersion',
    'StandardSecuritySettingsRevision',
    'PubKeyAdbeSubFilter',
    'CryptFilterConfiguration',
    'CryptFilter',
    'StandardCryptFilter',
    'PubKeyCryptFilter',
    'IdentityCryptFilter',
    'RC4CryptFilterMixin',
    'AESCryptFilterMixin',
    'StandardAESCryptFilter',
    'StandardRC4CryptFilter',
    'PubKeyAESCryptFilter',
    'PubKeyRC4CryptFilter',
    'EnvelopeKeyDecrypter',
    'SimpleEnvelopeKeyDecrypter',
    'SerialisedCredential',
    'SerialisableCredential',
    'STD_CF',
    'DEFAULT_CRYPT_FILTER',
    'DEF_EMBEDDED_FILE',
    'IDENTITY',
    'CryptFilterBuilder',
    'build_crypt_filter',
    'PdfKeyNotAvailableError',
]



import logging
import sys
from contextlib import contextmanager

import click
from pyhanko.cli.utils import logger
from pyhanko.config.logging import LogConfig, StdLogOutput
from pyhanko.pdf_utils import misc
from pyhanko.pdf_utils.layout import LayoutError
from pyhanko.sign.general import SigningError


class NoStackTraceFormatter(logging.Formatter):
    def formatException(self, ei) -> str:
        return ""  # pragma: nocover


LOG_FORMAT_STRING = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


def logging_setup(log_configs, verbose: bool):
    log_config: LogConfig
    for module, log_config in log_configs.items():
        cur_logger = logging.getLogger(module)
        cur_logger.setLevel(log_config.level)
        handler: logging.StreamHandler
        if isinstance(log_config.output, StdLogOutput):
            if StdLogOutput == StdLogOutput.STDOUT:
                handler = logging.StreamHandler(sys.stdout)
            else:
                handler = logging.StreamHandler()
            # when logging to the console, don't output stack traces
            # unless in verbose mode
            if verbose:
                formatter = logging.Formatter(LOG_FORMAT_STRING)
            else:
                formatter = NoStackTraceFormatter(LOG_FORMAT_STRING)
        else:
            handler = logging.FileHandler(log_config.output)
            formatter = logging.Formatter(LOG_FORMAT_STRING)
        handler.setFormatter(formatter)
        cur_logger.addHandler(handler)


@contextmanager
def pyhanko_exception_manager():
    msg = exception = None
    try:
        yield
    except click.ClickException:
        raise
    except misc.PdfStrictReadError as e:
        exception = e
        msg = (
            "Failed to read PDF file in strict mode; rerun with "
            "--no-strict-syntax to try again.\n"
            f"Error message: {e.msg}"
        )
    except misc.PdfReadError as e:
        exception = e
        msg = f"Failed to read PDF file: {e.msg}"
    except misc.PdfWriteError as e:
        exception = e
        msg = f"Failed to write PDF file: {e.msg}"
    except SigningError as e:
        exception = e
        msg = f"Error raised while producing signed file: {e.msg}"
    except LayoutError as e:
        exception = e
        msg = f"Error raised while producing signature layout: {e.msg}"
    except Exception as e:
        exception = e
        msg = "Generic processing error."

    if exception is not None:
        logger.error(msg, exc_info=exception)
        raise click.ClickException(msg)


DEFAULT_CONFIG_FILE = 'pyhanko.yml'



"""
Implementation of stream filters for PDF.

Taken from PyPDF2 with modifications. See :ref:`here <pypdf2-license>`
for the original license of the PyPDF2 project.

Note that not all decoders specified in the standard are supported.
In particular ``/LZWDecode`` and the various JPEG-based decoders are missing.
"""

import binascii
import re
import struct
import zlib
from io import BytesIO

from .crypt.api import SecurityHandler
from .misc import PdfStreamError, Singleton

__all__ = [
    'Decoder',
    'ASCII85Decode',
    'ASCIIHexDecode',
    'FlateDecode',
    'get_generic_decoder',
]

decompress = zlib.decompress
compress = zlib.compress


class Decoder:
    """
    General filter/decoder interface.
    """

    def decode(self, data: bytes, decode_params: dict) -> bytes:
        """
        Decode a stream.

        :param data:
            Data to decode.
        :param decode_params:
            Decoder parameters, sourced from the ``/DecoderParams`` entry
            associated with this filter.
        :return:
            Decoded data.
        """
        raise NotImplementedError

    def encode(self, data: bytes, decode_params: dict) -> bytes:
        """
        Encode a stream.

        :param data:
            Data to encode.
        :param decode_params:
            Encoder parameters, sourced from the ``/DecoderParams`` entry
            associated with this filter.
        :return:
            Encoded data.
        """
        raise NotImplementedError


def _png_decode(data: memoryview, columns):
    output = BytesIO()
    # PNG prediction can vary from row to row
    rowlength = columns + 1
    assert len(data) % rowlength == 0

    prev_result = bytes(rowlength - 1)
    for row in range(len(data) // rowlength):
        rowdata = data[(row * rowlength) : ((row + 1) * rowlength)]
        filter_byte = rowdata[0]
        result_row = bytearray(rowlength - 1)
        if filter_byte == 0:
            pass
        elif filter_byte == 1:
            pairs = zip(rowdata[2:], rowdata[1:])
            result_row[0] = rowdata[1]
            for i, (x, y) in enumerate(pairs):
                result_row[i + 1] = (x + y) % 256
        elif filter_byte == 2:
            pairs = zip(rowdata[1:], prev_result)
            for i, (x, y) in enumerate(pairs):
                result_row[i] = (x + y) % 256
        else:
            # unsupported PNG filter
            raise NotImplementedError("Unsupported PNG filter %r" % filter_byte)
        prev_result = result_row
        output.write(result_row)
    return output.getvalue()


class FlateDecode(Decoder, metaclass=Singleton):
    """
    Implementation of the ``/FlateDecode`` filter.

    .. warning::
        Currently not all predictor values are supported. This may cause
        problems when extracting image data from PDF files.
    """

    def decode(self, data: bytes, decode_params):
        # there's lots of slicing ahead, so let's reduce copying overhead
        data = memoryview(decompress(data))
        predictor = 1
        if decode_params:
            try:
                predictor = decode_params.get("/Predictor", 1)
            except AttributeError:
                pass  # usually an array with a null object was read

        # predictor 1 == no predictor
        if predictor == 1:
            return data

        columns = decode_params["/Columns"]
        # PNG prediction:
        if 10 <= predictor <= 15:
            return _png_decode(data, columns)
        else:
            # unsupported predictor
            raise NotImplementedError(
                "Unsupported flatedecode predictor %r" % predictor
            )

    def encode(self, data, decode_params=None):
        # TODO support the parameters in the spec
        return compress(data)


# TODO check boundary conditions in PDF spec

WS_REGEX = re.compile(b'\\s+')
ASCII_HEX_EOD_MARKER = b'>'


class ASCIIHexDecode(Decoder, metaclass=Singleton):
    """
    Wrapper around :func:`binascii.hexlify` that implements the
    :class:`.Decoder` interface.
    """

    def encode(self, data: bytes, decode_params=None) -> bytes:
        return binascii.hexlify(data) + b'>'

    def decode(self, data, decode_params=None):
        data, _ = data.split(ASCII_HEX_EOD_MARKER, 1)
        data = WS_REGEX.sub(b'', data)
        return binascii.unhexlify(data)


# TODO reimplement LZW decoder

ASCII_85_EOD_MARKER = b'~>'
POWS = tuple(85**p for p in (4, 3, 2, 1, 0))


class ASCII85Decode(Decoder, metaclass=Singleton):
    """
    Implementation of the base 85 encoding scheme specified in ISO 32000-1.
    """

    def encode(self, data: bytes, decode_params=None) -> bytes:
        # BytesIO is quite clever, in that it doesn't copy things until modified
        data_stm = BytesIO(data)
        out = BytesIO()

        while True:
            grp = data_stm.read(4)
            if not grp:
                break
            # This needs to happen before applying padding!
            # See § 7.4.3 in ISO 32000-1
            if grp == b'\0\0\0\0':
                out.write(b'z')
                continue

            bytes_read = len(grp)
            if bytes_read < 4:
                grp += b'\0' * (4 - bytes_read)
                pows = POWS[: bytes_read + 1]
            else:
                pows = POWS

            # write 5 chars in base85
            (grp_int,) = struct.unpack('>L', grp)
            for p in pows:
                digit, grp_int = divmod(grp_int, p)
                # use chars from 0x21 to 0x75
                out.write(bytes((digit + 0x21,)))
        out.write(ASCII_85_EOD_MARKER)
        return out.getvalue()

    def decode(self, data, decode_params=None):
        data, _ = data.split(ASCII_85_EOD_MARKER, 1)
        data_stm = BytesIO(WS_REGEX.sub(b'', data))
        out = BytesIO()
        while True:
            next_char = data_stm.read(1)
            if not next_char:
                break
            if next_char == b'z':
                out.write(b'\0\0\0\0')
                continue
            rest = data_stm.read(4)
            if not rest:  # pragma: nocover
                raise PdfStreamError(
                    'Nonzero ASCII85 group must have at least two digits.'
                )

            grp = next_char + rest
            grp_result = 0
            p = 0  # make the linter happy
            # convert back from base 85 to int
            for digit, p in zip(grp, POWS):
                digit -= 0x21
                if 0 <= digit < 85:
                    grp_result += p * digit
                else:  # pragma: nocover
                    raise PdfStreamError(
                        'Bytes in ASCII85 data must lie beteen 0x21 and 0x75.'
                    )
            # 85 and 256 are coprime, so the last digit will always be off by
            # one if we had to throw away a multiple of 256 in the encoding
            # step (due to padding).
            if len(grp) < 5:
                grp_result += p

            # Finally, pack the integer into a 4-byte unsigned int
            # (potentially need to cut off some excess digits)
            decoded = struct.pack('>L', grp_result)
            out.write(decoded[: len(grp) - 1])
        return out.getvalue()


class CryptFilterDecoder(Decoder):
    def __init__(self, handler: SecurityHandler):
        self.handler = handler

    def decode(self, data: bytes, decode_params: dict) -> bytes:
        from .crypt import IDENTITY

        cf_name = decode_params.get('/Name', IDENTITY)
        cf = self.handler.get_stream_filter(name=cf_name)
        # the spec explicitly tells us to use the global key here, go figure
        # (clause § 7.4.10 in both 32k-1 and 32k-2)
        return cf.decrypt(cf.shared_key, data, params=decode_params)

    def encode(self, data: bytes, decode_params: dict) -> bytes:
        from .crypt import IDENTITY

        cf_name = decode_params.get('/Name', IDENTITY)
        cf = self.handler.get_stream_filter(name=cf_name)
        return cf.encrypt(cf.shared_key, data, params=decode_params)


DECODERS = {
    '/FlateDecode': FlateDecode,
    '/Fl': FlateDecode,
    '/ASCIIHexDecode': ASCIIHexDecode,
    '/AHx': ASCIIHexDecode,
    '/ASCII85Decode': ASCII85Decode,
    '/A85': ASCII85Decode,
}


def get_generic_decoder(name: str) -> Decoder:
    """
    Instantiate a specific stream filter decoder type by (PDF) name.

    The following names are recognised:

    * ``/FlateDecode`` or ``/Fl`` for the decoder implementing Flate
       compression.
    * ``/ASCIIHexDecode`` or ``/AHx`` for the decoder that converts bytes to
      their hexadecimal representations.
    * ``/ASCII85Decode`` or ``/A85`` for the decoder that converts byte strings
      to a base-85 textual representation.

    .. warning::
        ``/Crypt`` is a special case because it requires access to the
        document's security handler.

    .. warning::
        LZW compression is currently unsupported, as are most compression
        methods that are used specifically for image data.

    :param name:
        Name of the decoder to instantiate.
    """

    try:
        cls = DECODERS[name]
    except KeyError:
        raise NotImplementedError(f"Stream filter '{name}' is not supported.")
    return cls()



import enum
import logging
from dataclasses import dataclass
from typing import Dict, Optional, Union

from pyhanko.config.errors import ConfigurationError
from pyhanko.pdf_utils.misc import get_and_apply

__all__ = ['StdLogOutput', 'LogConfig', 'parse_logging_config']


class StdLogOutput(enum.Enum):
    STDERR = enum.auto()
    STDOUT = enum.auto()


@dataclass(frozen=True)
class LogConfig:
    level: Union[int, str]
    """
    Logging level, should be one of the levels defined in the logging module.
    """

    output: Union[StdLogOutput, str]
    """
    Name of the output file, or a standard one.
    """

    @staticmethod
    def parse_output_spec(spec) -> Union[StdLogOutput, str]:
        if not isinstance(spec, str):
            raise ConfigurationError(
                "Log output must be specified as a string."
            )
        spec_l = spec.lower()
        if spec_l == 'stderr':
            return StdLogOutput.STDERR
        elif spec_l == 'stdout':
            return StdLogOutput.STDOUT
        else:
            return spec


DEFAULT_ROOT_LOGGER_LEVEL = logging.INFO


def _retrieve_log_level(settings_dict, key, default=None) -> Union[int, str]:
    try:
        level_spec = settings_dict[key]
    except KeyError:
        if default is not None:
            return default
        raise ConfigurationError(
            f"Logging config for '{key}' does not define a log level."
        )
    if not isinstance(level_spec, (int, str)):
        raise ConfigurationError(
            f"Log levels must be int or str, not {type(level_spec)}"
        )
    return level_spec


def parse_logging_config(log_config_spec) -> Dict[Optional[str], LogConfig]:
    if not isinstance(log_config_spec, dict):
        raise ConfigurationError('logging config should be a dictionary')

    root_logger_level = _retrieve_log_level(
        log_config_spec, 'root-level', default=DEFAULT_ROOT_LOGGER_LEVEL
    )

    root_logger_output = get_and_apply(
        log_config_spec,
        'root-output',
        LogConfig.parse_output_spec,
        default=StdLogOutput.STDERR,
    )

    log_config: Dict[Optional[str], LogConfig] = {
        None: LogConfig(root_logger_level, root_logger_output)
    }

    logging_by_module = log_config_spec.get('by-module', {})
    if not isinstance(logging_by_module, dict):
        raise ConfigurationError('logging.by-module should be a dict')

    for module, module_logging_settings in logging_by_module.items():
        if not isinstance(module, str):
            raise ConfigurationError(
                "Keys in logging.by-module should be strings"
            )
        level_spec = _retrieve_log_level(module_logging_settings, 'level')
        output_spec = get_and_apply(
            module_logging_settings,
            'output',
            LogConfig.parse_output_spec,
            default=StdLogOutput.STDERR,
        )
        log_config[module] = LogConfig(level=level_spec, output=output_spec)

    return log_config



import getpass
import os
from typing import ContextManager, List, Optional

import click
from pyhanko.cli._ctx import CLIContext
from pyhanko.cli.config import CLIConfig
from pyhanko.cli.plugin_api import SigningCommandPlugin
from pyhanko.cli.utils import logger, readable_file
from pyhanko.config.errors import ConfigurationError
from pyhanko.config.pkcs11 import (
    PKCS11PinEntryMode,
    PKCS11SignatureConfig,
    TokenCriteria,
)
from pyhanko.sign import Signer

__all__ = ['PKCS11Plugin']


try:
    import pkcs11  # lgtm [py/unused-import]

    pkcs11_available = True
except ImportError:  # pragma: nocover
    pkcs11 = None
    pkcs11_available = False


P11_PIN_ENV_VAR = "PYHANKO_PKCS11_PIN"
UNAVAIL_MSG = "This subcommand requires python-pkcs11 to be installed."


class PKCS11Plugin(SigningCommandPlugin):
    subcommand_name = 'pkcs11'
    help_summary = 'use generic PKCS#11 device to sign'
    unavailable_message = UNAVAIL_MSG

    def is_available(self) -> bool:
        return pkcs11_available

    def click_options(self) -> List[click.Option]:
        return [
            click.Option(
                ('--lib',),
                help='path to PKCS#11 module',
                type=readable_file,
                required=False,
            ),
            click.Option(
                ('--token-label',),
                help='PKCS#11 token label',
                type=str,
                required=False,
            ),
            click.Option(
                ('--cert-label',),
                help='certificate label',
                type=str,
                required=False,
            ),
            click.Option(
                ('--raw-mechanism',),
                help='invoke raw PKCS#11 mechanism',
                type=bool,
                is_flag=True,
                required=False,
            ),
            click.Option(
                ('--key-label',), help='key label', type=str, required=False
            ),
            click.Option(
                ('--slot-no',),
                help='specify PKCS#11 slot to use',
                required=False,
                type=int,
                default=None,
            ),
            click.Option(
                ('--skip-user-pin',),
                type=bool,
                show_default=True,
                default=False,
                required=False,
                is_flag=True,
                help='do not prompt for PIN (e.g. if the token has a PIN pad)',
            ),
            click.Option(
                ('--p11-setup',),
                type=str,
                required=False,
                help='name of preconfigured PKCS#11 profile (overrides all '
                'other options)',
            ),
            click.Option(
                ('--other-cert',),
                type=str,
                required=False,
                help='label of other cert to pull (multiple uses allowed)',
                multiple=True,
            ),
        ]

    def create_signer(
        self, context: CLIContext, **kwargs
    ) -> ContextManager[Signer]:
        return _pkcs11_signer_context(context, **kwargs)


def _pkcs11_signer_context(
    ctx: CLIContext,
    lib,
    token_label,
    cert_label,
    key_label,
    slot_no,
    skip_user_pin,
    p11_setup,
    raw_mechanism,
    other_cert,
):
    from pyhanko.sign import pkcs11

    if p11_setup:
        cli_config: Optional[CLIConfig] = ctx.config
        if cli_config is None:
            raise click.ClickException(
                "The --p11-setup option requires a configuration file"
            )
        try:
            pkcs11_config = ModuleConfigWrapper(cli_config).get_pkcs11_config(
                p11_setup
            )
        except ConfigurationError as e:
            msg = f"Error while reading PKCS#11 config {p11_setup}"
            logger.error(msg, exc_info=e)
            raise click.ClickException(msg)
    else:
        if not (lib and cert_label):
            raise click.ClickException(
                "The parameters --lib and --cert-label are required."
            )

        pinentry_mode = (
            PKCS11PinEntryMode.SKIP
            if skip_user_pin
            else PKCS11PinEntryMode.PROMPT
        )

        pkcs11_config = PKCS11SignatureConfig(
            module_path=lib,
            cert_label=cert_label,
            key_label=key_label,
            slot_no=slot_no,
            token_criteria=TokenCriteria(token_label),
            # for now, DEFER requires a config file
            prompt_pin=pinentry_mode,
            raw_mechanism=raw_mechanism,
            other_certs_to_pull=other_cert,
        )

    pin = pkcs11_config.user_pin

    # try to fetch the PIN from an env var
    if pin is None:
        pin_env = os.environ.get(P11_PIN_ENV_VAR, None)
        if pin_env:
            pin = pin_env.strip()

    if pkcs11_config.prompt_pin == PKCS11PinEntryMode.PROMPT and pin is None:
        pin = getpass.getpass(prompt='PKCS#11 user PIN: ')
    return pkcs11.PKCS11SigningContext(pkcs11_config, user_pin=pin)


class ModuleConfigWrapper:
    def __init__(self, config: CLIConfig):
        config_dict = config.raw_config
        self.pkcs11_setups = config_dict.get('pkcs11-setups', {})

    def get_pkcs11_config(self, name):
        try:
            setup = self.pkcs11_setups[name]
        except KeyError:
            raise ConfigurationError(f"There's no PKCS#11 setup named '{name}'")
        return PKCS11SignatureConfig.from_config(setup)



import click
from pyhanko.cli.commands.signing import signing
from pyhanko.cli.runtime import pyhanko_exception_manager
from pyhanko.cli.utils import parse_field_location_spec
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.pdf_utils.writer import copy_into_new_writer
from pyhanko.sign import fields

__all__ = ['list_sigfields', 'add_sig_field']


@signing.command(name='list', help='list signature fields')
@click.argument('infile', type=click.File('rb'))
@click.option(
    '--skip-status',
    help='do not print status',
    required=False,
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
def list_sigfields(infile, skip_status):
    with pyhanko_exception_manager():
        r = PdfFileReader(infile, strict=False)
        field_info = fields.enumerate_sig_fields(r)
        for ix, (name, value, field_ref) in enumerate(field_info):
            if skip_status:
                click.echo(name)
                continue
            click.echo(f"{name}:{'EMPTY' if value is None else 'FILLED'}")


@signing.command(
    name='addfields', help='add empty signature fields to a PDF field'
)
@click.argument('infile', type=click.File('rb'))
@click.argument('outfile', type=click.File('wb'))
@click.option(
    '--field',
    metavar='PAGE/X1,Y1,X2,Y2/NAME',
    multiple=True,
    required=True,
    help="Field specification (multiple allowed)",
)
@click.option(
    '--resave',
    is_flag=True,
    help="Resave the PDF document instead of creating an incremental update",
)
def add_sig_field(infile, outfile, field, resave):
    with pyhanko_exception_manager():
        if resave:
            writer = copy_into_new_writer(PdfFileReader(infile, strict=False))
        else:
            writer = IncrementalPdfFileWriter(infile, strict=False)

        for s in field:
            name, spec = parse_field_location_spec(s)
            assert spec is not None
            fields.append_signature_field(writer, spec)

        writer.write(outfile)
        infile.close()
        outfile.close()



import os
from fractions import Fraction
from io import BytesIO
from pathlib import Path

import pytest
from pyhanko.pdf_utils import generic, layout, writer
from pyhanko.pdf_utils.content import ImportedPdfPage, RawContent
from pyhanko.pdf_utils.font.opentype import GlyphAccumulatorFactory
from pyhanko.pdf_utils.images import PdfImage
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.layout import (
    AxisAlignment,
    LayoutError,
    Margins,
    SimpleBoxLayoutRule,
)
from pyhanko.pdf_utils.text import TextBoxStyle
from pyhanko.stamp import (
    STAMP_ART_CONTENT,
    CoordinateSystem,
    QRPosition,
    QRStamp,
    QRStampStyle,
    StaticStampStyle,
    TextStamp,
    TextStampStyle,
    qr_stamp_file,
    text_stamp_file,
)

from .layout_test_utils import compare_output, with_layout_comparison
from .samples import *

FONT_DIR = f'{TEST_DIR}/data/fonts'

# Noto fonts are licensed under the OFL
NOTO_SERIF_JP = f'{FONT_DIR}/NotoSerifJP-Regular.otf'
NOTO_SANS_ARABIC = f'{FONT_DIR}/NotoSansArabic-Regular.ttf'
NOTO_SANS = f'{FONT_DIR}/NotoSans-Regular.ttf'

# This font is licensed under a modified version of the GPLv3.
# See https://www.gnu.org/software/freefont/license.html
# We're only using it here because it's a useful example of a string-keyed
# CFF font.
FREE_SERIF = f'{FONT_DIR}/FreeSerif.otf'
EXPECTED_OUTPUT_DIR = f'{TEST_DIR}/data/pdf/layout-tests'


def test_simple_text_stamp(tmp_path):
    outfile: Path = tmp_path / "test-out.pdf"
    text_stamp_file(
        MINIMAL_PATH,
        str(outfile),
        TextStampStyle(stamp_text="Hi, it's\n%(ts)s"),
        dest_page=0,
        x=70,
        y=50,
    )


def test_simple_text_stamp_missing_params(tmp_path):
    outfile: Path = tmp_path / "test-out.pdf"
    msg = "Stamp text parameter 'foo' is missing"
    with pytest.raises(LayoutError, match=msg):
        text_stamp_file(
            MINIMAL_PATH,
            str(outfile),
            TextStampStyle(stamp_text="%(foo)s missing"),
            dest_page=0,
            x=70,
            y=50,
        )


def test_simple_qr_stamp(tmp_path):
    outfile: Path = tmp_path / "test-out.pdf"
    qr_stamp_file(
        MINIMAL_PATH,
        str(outfile),
        QRStampStyle(stamp_text="Hi, it's\n%(ts)s"),
        dest_page=0,
        x=70,
        y=50,
        url='https://example.com',
    )


def test_simple_qr_noto_stamp(tmp_path):
    outfile: Path = tmp_path / "test-out.pdf"

    ga_factory = GlyphAccumulatorFactory(NOTO_SERIF_JP)
    qr_stamp_file(
        MINIMAL_PATH,
        str(outfile),
        QRStampStyle(
            stamp_text="Hi, it's\n%(ts)s",
            text_box_style=TextBoxStyle(font=ga_factory),
        ),
        dest_page=0,
        x=70,
        y=50,
        url='https://example.com',
    )


def empty_page(stream_xrefs=False):
    w = writer.PdfFileWriter(stream_xrefs=stream_xrefs)
    page = writer.PageObject(
        contents=w.add_object(generic.StreamObject(stream_data=b'')),
        media_box=generic.ArrayObject([0, 0, 595, 842]),
    )
    w.insert_page(page)
    return w


def _arabic_text_page(stream_xrefs):
    w = empty_page(stream_xrefs=stream_xrefs)
    style = TextStampStyle(
        stamp_text='اَلْفُصْحَىٰ',
        text_box_style=TextBoxStyle(
            font=GlyphAccumulatorFactory(NOTO_SANS_ARABIC),
        ),
        inner_content_layout=layout.SimpleBoxLayoutRule(
            x_align=layout.AxisAlignment.ALIGN_MID,
            y_align=layout.AxisAlignment.ALIGN_MID,
            inner_content_scaling=layout.InnerScaling.STRETCH_TO_FIT,
            margins=layout.Margins.uniform(5),
        ),
    )
    ts = TextStamp(
        writer=w, style=style, box=layout.BoxConstraints(width=300, height=200)
    )
    ts.apply(0, x=10, y=60)
    return w


@with_layout_comparison
def test_arabic_box():
    w = _arabic_text_page(stream_xrefs=False)
    compare_output(
        writer=w, expected_output_path=f'{EXPECTED_OUTPUT_DIR}/arabic-box.pdf'
    )


@with_layout_comparison
def test_fonts_with_obj_streams():
    # this should automatically put some stuff in object streams
    w = _arabic_text_page(stream_xrefs=True)
    compare_output(
        writer=w, expected_output_path=f'{EXPECTED_OUTPUT_DIR}/arabic-box.pdf'
    )
    assert w.objs_in_streams


@pytest.mark.parametrize('stream_xrefs', [True, False])
@with_layout_comparison
def test_font_rewrite_idempotent(stream_xrefs):
    # There was a bug in the font embedding logic that would cause multiple
    # writes of the same file to fail.
    w = _arabic_text_page(stream_xrefs)
    out1 = BytesIO()
    w.write(out1)
    out2 = BytesIO()
    w.write(out2)
    assert abs(out1.seek(0, os.SEEK_END) - out2.seek(0, os.SEEK_END)) < 100


@with_layout_comparison
@pytest.mark.parametrize('fixed_size', [True, False])
def test_four_qr_stamps(fixed_size):
    # Share a font subset, the text is the same everywhere
    gaf = GlyphAccumulatorFactory(NOTO_SANS, font_size=10)
    w = empty_page()
    positions = ((10, 700), (10, 500), (10, 10), (260, 10))
    for qr_pos, (x, y) in zip(QRPosition, positions):
        style = QRStampStyle(
            stamp_text='Test stamp text\nAnother line of text',
            text_box_style=TextBoxStyle(font=gaf),
            qr_position=qr_pos,
            background=STAMP_ART_CONTENT,
            background_opacity=0.4,
        )
        if fixed_size:
            if qr_pos.horizontal_flow:
                box = layout.BoxConstraints(width=300, height=100)
            else:
                box = layout.BoxConstraints(width=100, height=300)
        else:
            box = None
        ts = QRStamp(writer=w, style=style, box=box, url='https://example.com')
        ts.apply(0, x=x, y=y)
    postfix = 'fixed' if fixed_size else 'natural'
    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/four-stamps-{postfix}.pdf')


@with_layout_comparison
def test_fancy_qr_stamp():
    w = empty_page()
    style = QRStampStyle(
        stamp_text='Test stamp text\nAnother line of text',
        qr_inner_content=STAMP_ART_CONTENT,
    )
    box = layout.BoxConstraints(width=300, height=100)
    QRStamp(writer=w, style=style, box=box, url='https://example.com').apply(
        0, x=10, y=500
    )

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/fancy-qr-stamp-test.pdf')


@with_layout_comparison
def test_japanese_vertical_text_stamp():
    gaf = GlyphAccumulatorFactory(
        NOTO_SERIF_JP,
        font_size=10,
        writing_direction='ttb',
        bcp47_lang_code='ja_JP',
    )
    w = empty_page()
    style = QRStampStyle(
        stamp_text=(
            'テスト\n縦書きテスト\n改行してみましょう（括弧）\nPDF\n'
            'ちょっと長めの文を書いてみた。'
        ),
        text_box_style=TextBoxStyle(font=gaf, vertical_text=True),
        qr_position=QRPosition.ABOVE_TEXT,
        background=STAMP_ART_CONTENT,
        background_opacity=0.4,
    )
    box = layout.BoxConstraints(width=100, height=300)
    ts = QRStamp(writer=w, style=style, box=box, url='https://example.com')
    ts.apply(0, x=10, y=415)
    ts = QRStamp(writer=w, style=style, box=None, url='https://example.com')
    ts.apply(0, x=400, y=415)
    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/ja-vert-stamps.pdf')


@with_layout_comparison
def test_stamp_with_unscaled_bitmap_bg():
    w = empty_page()

    long_text = '\n'.join(
        'Test test test test test test test test test '
        'on a bitmap background!'
        for _ in range(60)
    )
    style = TextStampStyle(
        stamp_text=long_text,
        background=PdfImage(f'{TEST_DIR}/data/img/stamp-indexed.png'),
    )

    ts = TextStamp(w, style)
    ts.apply(0, x=30, y=120)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/bitmap-bg.pdf')


@with_layout_comparison
def test_stamp_with_scaled_bitmap_bg():
    w = empty_page()

    text = '\n'.join(
        'Test test test test on a bitmap background!' for _ in range(3)
    )
    style = TextStampStyle(
        stamp_text=text,
        background=PdfImage(f'{TEST_DIR}/data/img/stamp-indexed.png'),
    )

    ts = TextStamp(w, style, box=layout.BoxConstraints(400, 100))
    ts.apply(0, x=30, y=600)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/scaled-bitmap-bg.pdf')


@with_layout_comparison
def test_stamp_with_undefined_bg_size():
    w = empty_page()

    undef_bg = RawContent(data=STAMP_ART_CONTENT.data)

    long_text = '\n'.join(
        'Test test test test test test test test test '
        'on an ill-defined background!'
        for _ in range(60)
    )
    style = TextStampStyle(stamp_text=long_text, background=undef_bg)

    ts = TextStamp(w, style)
    ts.apply(0, x=30, y=120)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/undef-bg-size.pdf')


@with_layout_comparison
def test_stamp_with_scaled_pdf_bg():
    w = empty_page()

    text = '\n'.join(
        'Test test test test on a PDF background!' for _ in range(3)
    )
    style = TextStampStyle(
        stamp_text=text,
        background=ImportedPdfPage(
            f'{TEST_DIR}/data/pdf/pdf-background-test.pdf'
        ),
    )

    ts = TextStamp(w, style, box=layout.BoxConstraints(200, 50))
    ts.apply(0, x=30, y=600)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/stamp-on-pdf-bg.pdf')


@with_layout_comparison
def test_stamp_with_fixed_pdf_content():
    w = empty_page()

    style = StaticStampStyle.from_pdf_file(
        f'{TEST_DIR}/data/pdf/pdf-background-test.pdf'
    )

    stamp = style.create_stamp(
        w, box=layout.BoxConstraints(200, 50), text_params={}
    )
    stamp.apply(0, x=30, y=600)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/stamp-from-static-pdf.pdf')


@pytest.mark.parametrize(
    'box',
    [
        None,
        layout.BoxConstraints(width=200),
        layout.BoxConstraints(height=50),
        layout.BoxConstraints(aspect_ratio=Fraction(4, 1)),
        layout.BoxConstraints(),
    ],
)
def test_static_stamp_enforce_box_defined(box):
    w = empty_page()

    style = StaticStampStyle.from_pdf_file(
        f'{TEST_DIR}/data/pdf/pdf-background-test.pdf'
    )

    with pytest.raises(layout.LayoutError, match="predetermined bounding box"):
        style.create_stamp(w, box=box, text_params={})


@with_layout_comparison
def test_simple_text_stamp_string_keyed_font():
    gaf = GlyphAccumulatorFactory(FREE_SERIF, font_size=10)

    w = empty_page()
    style = TextStampStyle(
        stamp_text=(
            "Hi, this is a string-keyed font test.\n"
            "This should have a ligature: difficult"
        ),
        text_box_style=TextBoxStyle(font=gaf),
    )

    ts = TextStamp(w, style, box=layout.BoxConstraints(200, 50))
    ts.apply(dest_page=0, x=70, y=50)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/freeserif-test.pdf')


@with_layout_comparison
def test_double_newline():
    gaf = GlyphAccumulatorFactory(FREE_SERIF, font_size=10)

    w = empty_page()
    style = TextStampStyle(
        stamp_text=(
            "Hi, this is a test. There should be two newlines here:\n\n"
            "And now a single newline:\n"
            "See?"
        ),
        text_box_style=TextBoxStyle(font=gaf),
    )

    ts = TextStamp(w, style, box=layout.BoxConstraints(200, 50))
    ts.apply(dest_page=0, x=70, y=50)

    compare_output(w, f'{EXPECTED_OUTPUT_DIR}/double-newline.pdf')


def test_zero_width_error():
    w = empty_page()
    zero_margins = SimpleBoxLayoutRule(
        x_align=AxisAlignment.ALIGN_MID,
        y_align=AxisAlignment.ALIGN_MID,
        margins=Margins(),
    )
    style = TextStampStyle(
        stamp_text='',
        border_width=0,
        inner_content_layout=zero_margins,
        text_box_style=TextBoxStyle(box_layout_rule=zero_margins),
    )
    box = layout.BoxConstraints(width=100, height=100)
    stamp = TextStamp(writer=w, style=style, box=box)

    inn_commands, (inn_width, inn_height) = stamp._inner_layout_natural_size()
    assert inn_width == 0

    # ...this shouldn't throw a division by zero error
    stamp.apply(0, x=10, y=500)


def test_simple_text_stamp_on_page_with_leaky_graphics_state():
    with open(f"{PDF_DATA_DIR}/leaky-graphics-state-doc.pdf", 'rb') as fin:
        pdf_out = IncrementalPdfFileWriter(fin, strict=False)
        stamp = TextStamp(
            writer=pdf_out, style=TextStampStyle(stamp_text="Hi, it's me")
        )
        stamp.apply(0, 70, 50)
        compare_output(
            pdf_out,
            f'{EXPECTED_OUTPUT_DIR}/leaky-graphics-state-stamp-result.pdf',
        )


def test_simple_text_stamp_on_page_with_leaky_graphics_state_without_coord_correction():
    with open(f"{PDF_DATA_DIR}/leaky-graphics-state-doc.pdf", 'rb') as fin:
        pdf_out = IncrementalPdfFileWriter(fin, strict=False)
        stamp = TextStamp(
            writer=pdf_out,
            style=TextStampStyle(stamp_text="Hi, it's me"),
        )
        stamp.apply(0, 70, 50, coords=CoordinateSystem.AMBIENT)
        compare_output(
            pdf_out,
            f'{EXPECTED_OUTPUT_DIR}/leaky-graphics-state-stamp-no-corr-result.pdf',
        )



import getpass

import click
from pyhanko.cli._root import cli_root
from pyhanko.cli.runtime import pyhanko_exception_manager
from pyhanko.cli.utils import _warn_empty_passphrase, readable_file
from pyhanko.keys import load_certs_from_pemder
from pyhanko.pdf_utils import crypt
from pyhanko.pdf_utils.crypt import StandardSecurityHandler
from pyhanko.pdf_utils.crypt.permissions import PubKeyPermissions
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.pdf_utils.writer import copy_into_new_writer

__all__ = ['decrypt', 'encrypt_file']


@cli_root.command(help='encrypt PDF files (AES-256 only)', name='encrypt')
@click.argument('infile', type=readable_file)
@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
@click.option(
    '--password',
    help='password to encrypt the file with',
    required=False,
    type=str,
)
@click.option(
    '--recipient',
    required=False,
    multiple=True,
    help='certificate(s) corresponding to entities that '
    'can decrypt the output file',
    type=click.Path(readable=True, dir_okay=False),
)
def encrypt_file(infile, outfile, password, recipient):
    if password and recipient:
        raise click.ClickException(
            "Specify either a password or a list of recipients."
        )
    elif not password and not recipient:
        password = getpass.getpass(prompt='Output file password: ')

    recipient_certs = None
    if recipient:
        recipient_certs = list(load_certs_from_pemder(cert_files=recipient))

    with pyhanko_exception_manager():
        with open(infile, 'rb') as inf:
            r = PdfFileReader(inf)
            w = copy_into_new_writer(r)

            if recipient_certs:
                w.encrypt_pubkey(recipient_certs)
            else:
                w.encrypt(owner_pass=password)

            with open(outfile, 'wb') as outf:
                w.write(outf)


@cli_root.group(
    help='decrypt PDF files (any standard PDF encryption scheme)',
    name='decrypt',
)
def decrypt():
    pass


decrypt_force_flag = click.option(
    '--force',
    help='ignore access restrictions (use at your own risk)',
    required=False,
    type=bool,
    is_flag=True,
    default=False,
)


@decrypt.command(help='decrypt using password', name='password')
@click.argument('infile', type=readable_file)
@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
@click.option(
    '--password',
    help='password to decrypt the file with',
    required=False,
    type=str,
)
@decrypt_force_flag
def decrypt_with_password(infile, outfile, password, force):
    with pyhanko_exception_manager():
        with open(infile, 'rb') as inf:
            r = PdfFileReader(inf)
            if r.security_handler is None:
                raise click.ClickException("File is not encrypted.")
            elif not isinstance(r.security_handler, StandardSecurityHandler):
                raise click.ClickException(
                    "File is not encrypted with the standard (password-based) security handler"
                )
            if not password:
                password = getpass.getpass(prompt='File password: ')
            auth_result = r.decrypt(password)
            if auth_result.status == crypt.AuthStatus.USER and not force:
                raise click.ClickException(
                    "Password specified was the user password, not "
                    "the owner password. Pass --force to decrypt the "
                    "file anyway."
                )
            elif auth_result.status == crypt.AuthStatus.FAILED:
                raise click.ClickException("Password didn't match.")
            w = copy_into_new_writer(r)
            with open(outfile, 'wb') as outf:
                w.write(outf)


@decrypt.command(help='decrypt using private key (PEM/DER)', name='pemder')
@click.argument('infile', type=readable_file)
@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
@click.option(
    '--key',
    type=readable_file,
    required=True,
    help='file containing the recipient\'s private key (PEM/DER)',
)
@click.option(
    '--cert',
    help='file containing the recipient\'s certificate (PEM/DER)',
    type=readable_file,
    required=True,
)
@click.option(
    '--passfile',
    required=False,
    type=click.File('rb'),
    help='file containing the passphrase for the private key',
    show_default='stdin',
)
@click.option(
    '--no-pass',
    help='assume the private key file is unencrypted',
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
)
@decrypt_force_flag
def decrypt_with_pemder(infile, outfile, key, cert, passfile, force, no_pass):
    if passfile is not None:
        passphrase = passfile.read()
        passfile.close()
    elif not no_pass:
        passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
        if not passphrase:
            _warn_empty_passphrase()
            passphrase = None
    else:
        passphrase = None

    sedk = crypt.SimpleEnvelopeKeyDecrypter.load(
        key, cert, key_passphrase=passphrase
    )

    _decrypt_pubkey(sedk, infile, outfile, force)


def _decrypt_pubkey(
    sedk: crypt.SimpleEnvelopeKeyDecrypter, infile, outfile, force
):
    with pyhanko_exception_manager():
        with open(infile, 'rb') as inf:
            r = PdfFileReader(inf)
            if r.security_handler is None:
                raise click.ClickException("File is not encrypted.")
            if not isinstance(r.security_handler, crypt.PubKeySecurityHandler):
                raise click.ClickException(
                    "File was not encrypted with a public-key security handler."
                )
            auth_result = r.decrypt_pubkey(sedk)
            if auth_result.status == crypt.AuthStatus.USER:
                if (
                    not force
                    and auth_result.permission_flags
                    and not (
                        PubKeyPermissions.ALLOW_ENCRYPTION_CHANGE
                        in auth_result.permission_flags
                    )
                ):
                    raise click.ClickException(
                        "Change of encryption is typically not allowed with "
                        "user access. Pass --force to decrypt the file anyway."
                    )
            elif auth_result.status == crypt.AuthStatus.FAILED:
                raise click.ClickException("Failed to decrypt the file.")
            w = copy_into_new_writer(r)
            with open(outfile, 'wb') as outf:
                w.write(outf)


@decrypt.command(help='decrypt using private key (PKCS#12)', name='pkcs12')
@click.argument('infile', type=readable_file)
@click.argument('outfile', type=click.Path(writable=True, dir_okay=False))
@click.argument('pfx', type=readable_file)
@click.option(
    '--passfile',
    required=False,
    type=click.File('r'),
    help='file containing the passphrase for the PKCS#12 file',
    show_default='stdin',
)
@decrypt_force_flag
def decrypt_with_pkcs12(infile, outfile, pfx, passfile, force):
    if passfile is None:
        passphrase = getpass.getpass(prompt='Key passphrase: ').encode('utf-8')
    else:
        passphrase = passfile.readline().strip().encode('utf-8')
        passfile.close()
    sedk = crypt.SimpleEnvelopeKeyDecrypter.load_pkcs12(
        pfx, passphrase=passphrase
    )

    _decrypt_pubkey(sedk, infile, outfile, force)
